# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.EbuildMetadataPhase import EbuildMetadataPhase

import portage
from portage import os
from portage.cache.cache_errors import CacheError
from portage.dep import _repo_separator
from portage.util._async.AsyncScheduler import AsyncScheduler


class MetadataRegen(AsyncScheduler):
    def __init__(self, portdb, cp_iter=None, consumer=None, write_auxdb=True, **kwargs):
        AsyncScheduler.__init__(self, **kwargs)
        self._portdb = portdb
        self._write_auxdb = write_auxdb
        self._global_cleanse = False
        if cp_iter is None:
            cp_iter = self._iter_every_cp()
            # We can globally cleanse stale cache only if we
            # iterate over every single cp.
            self._global_cleanse = True
        self._cp_iter = cp_iter
        self._consumer = consumer

        self._valid_pkgs = set()
        self._cp_set = set()
        self._process_iter = self._iter_metadata_processes()
        self._running_tasks = set()

    def _next_task(self):
        return next(self._process_iter)

    def _iter_every_cp(self):
        # List categories individually, in order to start yielding quicker,
        # and in order to reduce latency in case of a signal interrupt.
        cp_all = self._portdb.cp_all
        for category in sorted(self._portdb.categories):
            yield from cp_all(categories=(category,))

    def _iter_metadata_processes(self):
        portdb = self._portdb
        valid_pkgs = self._valid_pkgs
        cp_set = self._cp_set
        consumer = self._consumer

        portage.writemsg_stdout("Regenerating cache entries...\n")
        for cp in self._cp_iter:
            if self._terminated.is_set():
                break
            cp_set.add(cp)
            portage.writemsg_stdout("Processing %s\n" % cp)
            # We iterate over portdb.porttrees, since it's common to
            # tweak this attribute in order to adjust repo selection.
            for mytree in portdb.porttrees:
                repo = portdb.repositories.get_repo_for_location(mytree)
                cpv_list = portdb.cp_list(cp, mytree=[repo.location])
                for cpv in cpv_list:
                    if self._terminated.is_set():
                        break
                    valid_pkgs.add(cpv)
                    ebuild_path, repo_path = portdb.findname2(cpv, myrepo=repo.name)
                    if ebuild_path is None:
                        raise AssertionError(
                            "ebuild not found for '%s%s%s'"
                            % (cpv, _repo_separator, repo.name)
                        )
                    metadata, ebuild_hash = portdb._pull_valid_cache(
                        cpv, ebuild_path, repo_path
                    )
                    if metadata is not None:
                        if consumer is not None:
                            consumer(cpv, repo_path, metadata, ebuild_hash, True)
                        continue

                    yield EbuildMetadataPhase(
                        cpv=cpv,
                        ebuild_hash=ebuild_hash,
                        portdb=portdb,
                        repo_path=repo_path,
                        settings=portdb.doebuild_settings,
                        write_auxdb=self._write_auxdb,
                    )

    def _cleanup(self):
        super()._cleanup()

        portdb = self._portdb
        dead_nodes = {}

        if self._terminated.is_set():
            portdb.flush_cache()
            return

        if self._global_cleanse:
            for mytree in portdb.porttrees:
                try:
                    dead_nodes[mytree] = set(portdb.auxdb[mytree])
                except CacheError as e:
                    portage.writemsg(
                        "Error listing cache entries for "
                        + "'{}': {}, continuing...\n".format(mytree, e),
                        noiselevel=-1,
                    )
                    del e
                    dead_nodes = None
                    break
        else:
            cp_set = self._cp_set
            cpv_getkey = portage.cpv_getkey
            for mytree in portdb.porttrees:
                try:
                    dead_nodes[mytree] = {
                        cpv for cpv in portdb.auxdb[mytree] if cpv_getkey(cpv) in cp_set
                    }
                except CacheError as e:
                    portage.writemsg(
                        "Error listing cache entries for "
                        + "'{}': {}, continuing...\n".format(mytree, e),
                        noiselevel=-1,
                    )
                    del e
                    dead_nodes = None
                    break

        if dead_nodes:
            for y in self._valid_pkgs:
                for mytree in portdb.porttrees:
                    if portdb.findname2(y, mytree=mytree)[0]:
                        dead_nodes[mytree].discard(y)

            for mytree, nodes in dead_nodes.items():
                auxdb = portdb.auxdb[mytree]
                for y in nodes:
                    try:
                        del auxdb[y]
                    except (KeyError, CacheError):
                        pass

        portdb.flush_cache()

    def _task_exit(self, metadata_process):

        if metadata_process.returncode != os.EX_OK:
            self._valid_pkgs.discard(metadata_process.cpv)
            if not self._terminated_tasks:
                portage.writemsg(
                    "Error processing {}, continuing...\n".format(metadata_process.cpv),
                    noiselevel=-1,
                )

        if self._consumer is not None:
            # On failure, still notify the consumer (in this case the metadata
            # argument is None).
            self._consumer(
                metadata_process.cpv,
                metadata_process.repo_path,
                metadata_process.metadata,
                metadata_process.ebuild_hash,
                metadata_process.eapi_supported,
            )

        AsyncScheduler._task_exit(self, metadata_process)



# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import get_required_use_flags
from portage.exception import InvalidDependString


class TestCheckRequiredUse(TestCase):
    def testCheckRequiredUse(self):
        test_cases = (
            ("a b c", ["a", "b", "c"]),
            ("|| ( a b c )", ["a", "b", "c"]),
            ("^^ ( a b c )", ["a", "b", "c"]),
            ("?? ( a b c )", ["a", "b", "c"]),
            ("?? ( )", []),
            ("|| ( a b ^^ ( d e f ) )", ["a", "b", "d", "e", "f"]),
            ("^^ ( a b || ( d e f ) )", ["a", "b", "d", "e", "f"]),
            ("( ^^ ( a ( b ) ( || ( ( d e ) ( f ) ) ) ) )", ["a", "b", "d", "e", "f"]),
            ("a? ( ^^ ( b c ) )", ["a", "b", "c"]),
            ("a? ( ^^ ( !b !d? ( c ) ) )", ["a", "b", "c", "d"]),
        )

        test_cases_xfail = (
            ("^^ ( || ( a b ) ^^ ( b c )"),
            ("^^( || ( a b ) ^^ ( b c ) )"),
            ("^^ || ( a b ) ^^ ( b c )"),
            ("^^ ( ( || ) ( a b ) ^^ ( b c ) )"),
            ("^^ ( || ( a b ) ) ^^ ( b c ) )"),
        )

        for required_use, expected in test_cases:
            result = get_required_use_flags(required_use)
            expected = set(expected)
            self.assertEqual(
                result,
                expected,
                "REQUIRED_USE: '%s', expected: '%s', got: '%s'"
                % (required_use, expected, result),
            )

        for required_use in test_cases_xfail:
            self.assertRaisesMsg(
                "REQUIRED_USE: '{}'".format(required_use),
                InvalidDependString,
                get_required_use_flags,
                required_use,
            )



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools
import logging
import portage
from portage import os
from portage.dbapi._MergeProcess import MergeProcess
from portage.exception import UnsupportedAPIException
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from _emerge.EbuildBuildDir import EbuildBuildDir
from _emerge.emergelog import emergelog
from _emerge.CompositeTask import CompositeTask
from _emerge.unmerge import _unmerge_display


class PackageUninstall(CompositeTask):
    """
    Uninstall a package asynchronously in a subprocess. When
    both parallel-install and ebuild-locks FEATURES are enabled,
    it is essential for the ebuild-locks code to execute in a
    subprocess, since the portage.locks module does not behave
    as desired if we try to lock the same file multiple times
    concurrently from the same process for ebuild-locks phases
    such as pkg_setup, pkg_prerm, and pkg_postrm.
    """

    __slots__ = (
        "world_atom",
        "ldpath_mtimes",
        "opts",
        "pkg",
        "settings",
        "_builddir_lock",
    )

    def _start(self):

        vardb = self.pkg.root_config.trees["vartree"].dbapi
        dbdir = vardb.getpath(self.pkg.cpv)
        if not os.path.exists(dbdir):
            # Apparently the package got uninstalled
            # already, so we can safely return early.
            self.returncode = os.EX_OK
            self._async_wait()
            return

        self.settings.setcpv(self.pkg)
        cat, pf = portage.catsplit(self.pkg.cpv)
        myebuildpath = os.path.join(dbdir, pf + ".ebuild")

        try:
            portage.doebuild_environment(
                myebuildpath, "prerm", settings=self.settings, db=vardb
            )
        except UnsupportedAPIException:
            # This is safe to ignore since this function is
            # guaranteed to set PORTAGE_BUILDDIR even though
            # it raises UnsupportedAPIException. The error
            # will be logged when it prevents the pkg_prerm
            # and pkg_postrm phases from executing.
            pass

        self._builddir_lock = EbuildBuildDir(
            scheduler=self.scheduler, settings=self.settings
        )
        self._start_task(
            AsyncTaskFuture(future=self._builddir_lock.async_lock()),
            self._start_unmerge,
        )

    def _start_unmerge(self, lock_task):
        self._assert_current(lock_task)
        if lock_task.cancelled:
            self._default_final_exit(lock_task)
            return

        lock_task.future.result()
        portage.prepare_build_dirs(settings=self.settings, cleanup=True)

        # Output only gets logged if it comes after prepare_build_dirs()
        # which initializes PORTAGE_LOG_FILE.
        retval, _ = _unmerge_display(
            self.pkg.root_config,
            self.opts,
            "unmerge",
            [self.pkg.cpv],
            clean_delay=0,
            writemsg_level=self._writemsg_level,
        )

        if retval != os.EX_OK:
            self._async_unlock_builddir(returncode=retval)
            return

        self._writemsg_level(
            ">>> Unmerging {}...\n".format(self.pkg.cpv), noiselevel=-1
        )
        self._emergelog("=== Unmerging... ({})".format(self.pkg.cpv))

        cat, pf = portage.catsplit(self.pkg.cpv)
        unmerge_task = MergeProcess(
            mycat=cat,
            mypkg=pf,
            settings=self.settings,
            treetype="vartree",
            vartree=self.pkg.root_config.trees["vartree"],
            scheduler=self.scheduler,
            background=self.background,
            mydbapi=self.pkg.root_config.trees["vartree"].dbapi,
            prev_mtimes=self.ldpath_mtimes,
            logfile=self.settings.get("PORTAGE_LOG_FILE"),
            unmerge=True,
        )

        self._start_task(unmerge_task, self._unmerge_exit)

    def _unmerge_exit(self, unmerge_task):
        if self._final_exit(unmerge_task) != os.EX_OK:
            self._emergelog(" !!! unmerge FAILURE: {}".format(self.pkg.cpv))
        else:
            self._emergelog(" >>> unmerge success: {}".format(self.pkg.cpv))
            self.world_atom(self.pkg)
        self._async_unlock_builddir(returncode=self.returncode)

    def _async_unlock_builddir(self, returncode=None):
        """
        Release the lock asynchronously, and if a returncode parameter
        is given then set self.returncode and notify exit listeners.
        """
        if returncode is not None:
            # The returncode will be set after unlock is complete.
            self.returncode = None
        self._start_task(
            AsyncTaskFuture(future=self._builddir_lock.async_unlock()),
            functools.partial(self._unlock_builddir_exit, returncode=returncode),
        )

    def _unlock_builddir_exit(self, unlock_task, returncode=None):
        self._assert_current(unlock_task)
        if unlock_task.cancelled and returncode is not None:
            self._default_final_exit(unlock_task)
            return

        # Normally, async_unlock should not raise an exception here.
        unlock_task.future.cancelled() or unlock_task.future.result()
        if returncode is not None:
            self.returncode = returncode
            self._async_wait()

    def _emergelog(self, msg):
        emergelog("notitles" not in self.settings.features, msg)

    def _writemsg_level(self, msg, level=0, noiselevel=0):

        log_path = self.settings.get("PORTAGE_LOG_FILE")
        background = self.background

        if log_path is None:
            if not (background and level < logging.WARNING):
                portage.util.writemsg_level(msg, level=level, noiselevel=noiselevel)
        else:
            self.scheduler.output(
                msg, log_path=log_path, level=level, noiselevel=noiselevel
            )



# Copyright 2010-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.AbstractEbuildProcess import AbstractEbuildProcess


class EbuildSpawnProcess(AbstractEbuildProcess):
    """
    Used by doebuild.spawn() to manage the spawned process.
    """

    _spawn_kwarg_names = AbstractEbuildProcess._spawn_kwarg_names + ("fakeroot_state",)

    __slots__ = ("fakeroot_state", "spawn_func")

    def _spawn(self, args, **kwargs):

        env = self.settings.environ()

        if self._dummy_pipe_fd is not None:
            env["PORTAGE_PIPE_FD"] = str(self._dummy_pipe_fd)

        return self.spawn_func(args, env=env, **kwargs)



# Copyright 2011-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tempfile
import traceback

import portage
from portage import os
from portage import shutil
from portage.exception import TryAgain
from portage.tests import TestCase


class LockNonblockTestCase(TestCase):
    def _testLockNonblock(self):
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = portage.locks.lockfile(path)
            pid = os.fork()
            if pid == 0:
                portage._ForkWatcher.hook(portage._ForkWatcher)
                portage.locks._close_fds()
                # Disable close_fds since we don't exec
                # (see _setup_pipes docstring).
                portage.process._setup_pipes({0: 0, 1: 1, 2: 2}, close_fds=False)
                rval = 2
                try:
                    try:
                        lock2 = portage.locks.lockfile(path, flags=os.O_NONBLOCK)
                    except portage.exception.TryAgain:
                        rval = os.EX_OK
                    else:
                        rval = 1
                        portage.locks.unlockfile(lock2)
                except SystemExit:
                    raise
                except:
                    traceback.print_exc()
                finally:
                    os._exit(rval)

            self.assertEqual(pid > 0, True)
            pid, status = os.waitpid(pid, 0)
            self.assertEqual(os.WIFEXITED(status), True)
            self.assertEqual(os.WEXITSTATUS(status), os.EX_OK)

            portage.locks.unlockfile(lock1)
        finally:
            shutil.rmtree(tempdir)

    def testLockNonblock(self):
        self._testLockNonblock()

    def testLockNonblockHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testLockNonblock()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def test_competition_with_same_process(self):
        """
        Test that at attempt to lock the same file multiple times in the
        same process will behave as intended (bug 714480).
        """
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock = portage.locks.lockfile(path)
            self.assertRaises(
                TryAgain, portage.locks.lockfile, path, flags=os.O_NONBLOCK
            )
            portage.locks.unlockfile(lock)
        finally:
            shutil.rmtree(tempdir)



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class AutounmaskMultilibUseTestCase(TestCase):
    def testAutounmaskMultilibUse(self):

        self.todo = True

        ebuilds = {
            "x11-proto/xextproto-7.2.1-r1": {
                "EAPI": "5",
                "IUSE": "abi_x86_32 abi_x86_64",
            },
            "x11-libs/libXaw-1.0.11-r2": {
                "EAPI": "5",
                "IUSE": "abi_x86_32 abi_x86_64",
                "RDEPEND": "x11-proto/xextproto[abi_x86_32(-)?,abi_x86_64(-)?]",
            },
            "app-emulation/emul-linux-x86-xlibs-20130224-r2": {
                "EAPI": "5",
                "RDEPEND": "x11-libs/libXaw[abi_x86_32]",
            },
            "games-util/steam-client-meta-0-r20130514": {
                "EAPI": "5",
                "RDEPEND": "app-emulation/emul-linux-x86-xlibs",
            },
        }

        installed = {
            "x11-proto/xextproto-7.2.1-r1": {
                "EAPI": "5",
                "IUSE": "abi_x86_32 abi_x86_64",
                "USE": "abi_x86_32 abi_x86_64",
            },
            "x11-libs/libXaw-1.0.11-r2": {
                "EAPI": "5",
                "IUSE": "abi_x86_32 abi_x86_64",
                "RDEPEND": "x11-proto/xextproto[abi_x86_32(-)?,abi_x86_64(-)?]",
                "USE": "abi_x86_32 abi_x86_64",
            },
            "app-emulation/emul-linux-x86-xlibs-20130224-r2": {
                "EAPI": "5",
                "RDEPEND": "x11-libs/libXaw[abi_x86_32]",
            },
            "games-util/steam-client-meta-0-r20130514": {
                "EAPI": "5",
                "RDEPEND": "app-emulation/emul-linux-x86-xlibs",
            },
        }

        user_config = {
            # "make.conf" : ("USE=\"abi_x86_32 abi_x86_64\"",)
            "make.conf": ('USE="abi_x86_64"',)
        }

        world = ("games-util/steam-client-meta",)

        test_cases = (
            # Test autounmask solving of multilib use deps for bug #481628.
            # We would like it to suggest some USE changes, but instead it
            # currently fails with a SLOT conflict.
            ResolverPlaygroundTestCase(
                ["x11-proto/xextproto", "x11-libs/libXaw"],
                options={"--oneshot": True, "--autounmask": True, "--backtrack": 30},
                mergelist=["x11-proto/xextproto-7.2.1-r1", "x11-libs/libXaw-1.0.11-r2"],
                success=True,
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            user_config=user_config,
            world=world,
            debug=False,
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# elog/__init__.py - elog core functions
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.util:writemsg",
)

from portage.const import EBUILD_PHASES
from portage.exception import AlarmSignal, PortageException
from portage.process import atexit_register
from portage.elog.messages import collect_ebuild_messages, collect_messages
from portage.elog.filtering import filter_loglevels
from portage.localization import _
from portage import os


def _preload_elog_modules(settings):
    logsystems = settings.get("PORTAGE_ELOG_SYSTEM", "").split()
    for s in logsystems:
        # allow per module overrides of PORTAGE_ELOG_CLASSES
        if ":" in s:
            s, levels = s.split(":", 1)
            levels = levels.split(",")
        # - is nicer than _ for module names, so allow people to use it.
        s = s.replace("-", "_")
        try:
            _load_mod("portage.elog.mod_" + s)
        except ImportError:
            pass


def _merge_logentries(a, b):
    rValue = {}
    phases = set(a)
    phases.update(b)
    for p in phases:
        merged_msgs = []
        rValue[p] = merged_msgs
        for d in a, b:
            msgs = d.get(p)
            if msgs:
                merged_msgs.extend(msgs)
    return rValue


def _combine_logentries(logentries):
    # generate a single string with all log messages
    rValue = []
    for phase in EBUILD_PHASES:
        if not phase in logentries:
            continue
        previous_type = None
        for msgtype, msgcontent in logentries[phase]:
            if previous_type != msgtype:
                previous_type = msgtype
                rValue.append("{}: {}".format(msgtype, phase))
            if isinstance(msgcontent, str):
                rValue.append(msgcontent.rstrip("\n"))
            else:
                for line in msgcontent:
                    rValue.append(line.rstrip("\n"))
    if rValue:
        rValue.append("")
    return "\n".join(rValue)


_elog_mod_imports = {}


def _load_mod(name):
    global _elog_mod_imports
    m = _elog_mod_imports.get(name)
    if m is None:
        m = __import__(name)
        for comp in name.split(".")[1:]:
            m = getattr(m, comp)
        _elog_mod_imports[name] = m
    return m


_elog_listeners = []


def add_listener(listener):
    """
    Listeners should accept four arguments: settings, key, logentries and logtext
    """
    _elog_listeners.append(listener)


def remove_listener(listener):
    """
    Remove previously added listener
    """
    _elog_listeners.remove(listener)


_elog_atexit_handlers = []


def elog_process(cpv, mysettings, phasefilter=None):
    global _elog_atexit_handlers

    logsystems = mysettings.get("PORTAGE_ELOG_SYSTEM", "").split()
    for s in logsystems:
        # allow per module overrides of PORTAGE_ELOG_CLASSES
        if ":" in s:
            s, levels = s.split(":", 1)
            levels = levels.split(",")
        # - is nicer than _ for module names, so allow people to use it.
        s = s.replace("-", "_")
        try:
            _load_mod("portage.elog.mod_" + s)
        except ImportError:
            pass

    if "T" in mysettings:
        ebuild_logentries = collect_ebuild_messages(
            os.path.join(mysettings["T"], "logging")
        )
    else:
        # A build dir isn't necessarily required since the messages.e*
        # functions allow messages to be generated in-memory.
        ebuild_logentries = {}
    all_logentries = collect_messages(key=cpv, phasefilter=phasefilter)
    if cpv in all_logentries:
        # Messages generated by the python elog implementation are assumed
        # to come first. For example, this ensures correct order for einfo
        # messages that are generated prior to the setup phase.
        all_logentries[cpv] = _merge_logentries(all_logentries[cpv], ebuild_logentries)
    else:
        all_logentries[cpv] = ebuild_logentries

    my_elog_classes = set(mysettings.get("PORTAGE_ELOG_CLASSES", "").split())
    logsystems = {}
    for token in mysettings.get("PORTAGE_ELOG_SYSTEM", "").split():
        if ":" in token:
            s, levels = token.split(":", 1)
            levels = levels.split(",")
        else:
            s = token
            levels = ()
        levels_set = logsystems.get(s)
        if levels_set is None:
            levels_set = set()
            logsystems[s] = levels_set
        levels_set.update(levels)

    for key in all_logentries:
        default_logentries = filter_loglevels(all_logentries[key], my_elog_classes)

        # in case the filters matched all messages and no module overrides exist
        if len(default_logentries) == 0 and (
            not ":" in mysettings.get("PORTAGE_ELOG_SYSTEM", "")
        ):
            continue

        default_fulllog = _combine_logentries(default_logentries)

        # call listeners
        for listener in _elog_listeners:
            listener(mysettings, str(key), default_logentries, default_fulllog)

        # pass the processing to the individual modules
        for s, levels in logsystems.items():
            # allow per module overrides of PORTAGE_ELOG_CLASSES
            if levels:
                mod_logentries = filter_loglevels(all_logentries[key], levels)
                mod_fulllog = _combine_logentries(mod_logentries)
            else:
                mod_logentries = default_logentries
                mod_fulllog = default_fulllog
            if len(mod_logentries) == 0:
                continue
            # - is nicer than _ for module names, so allow people to use it.
            s = s.replace("-", "_")
            try:
                m = _load_mod("portage.elog.mod_" + s)
                # Timeout after one minute (in case something like the mail
                # module gets hung).
                try:
                    AlarmSignal.register(60)
                    m.process(mysettings, str(key), mod_logentries, mod_fulllog)
                finally:
                    AlarmSignal.unregister()
                if hasattr(m, "finalize") and not m.finalize in _elog_atexit_handlers:
                    _elog_atexit_handlers.append(m.finalize)
                    atexit_register(m.finalize)
            except (ImportError, AttributeError) as e:
                writemsg(
                    _(
                        "!!! Error while importing logging modules "
                        'while loading "mod_%s":\n'
                    )
                    % str(s)
                )
                writemsg("%s\n" % str(e), noiselevel=-1)
            except AlarmSignal:
                writemsg("Timeout in elog_process for system '%s'\n" % s, noiselevel=-1)
            except PortageException as e:
                writemsg("%s\n" % str(e), noiselevel=-1)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging
import signal
import sys
import textwrap
import portage
from portage import os
from portage.dbapi._expand_new_virt import expand_new_virt
from portage.localization import _
from portage.output import bold, colorize, darkgreen, green
from portage._sets import SETPREFIX
from portage._sets.base import EditablePackageSet
from portage.versions import cpv_sort_key, _pkg_str

from _emerge.emergelog import emergelog
from _emerge.Package import Package
from _emerge.UserQuery import UserQuery
from _emerge.UninstallFailure import UninstallFailure
from _emerge.countdown import countdown


def _unmerge_display(
    root_config,
    myopts,
    unmerge_action,
    unmerge_files,
    clean_delay=1,
    ordered=0,
    writemsg_level=portage.util.writemsg_level,
):
    """
    Returns a tuple of (returncode, pkgmap) where returncode is
    os.EX_OK if no errors occur, and 1 otherwise.
    """

    quiet = "--quiet" in myopts
    settings = root_config.settings
    sets = root_config.sets
    vartree = root_config.trees["vartree"]
    candidate_catpkgs = []
    global_unmerge = 0
    out = portage.output.EOutput()
    pkg_cache = {}
    db_keys = list(vartree.dbapi._aux_cache_keys)

    def _pkg(cpv):
        pkg = pkg_cache.get(cpv)
        if pkg is None:
            pkg = Package(
                built=True,
                cpv=cpv,
                installed=True,
                metadata=zip(db_keys, vartree.dbapi.aux_get(cpv, db_keys)),
                operation="uninstall",
                root_config=root_config,
                type_name="installed",
            )
            pkg_cache[cpv] = pkg
        return pkg

    vdb_path = os.path.join(settings["EROOT"], portage.VDB_PATH)
    try:
        # At least the parent needs to exist for the lock file.
        portage.util.ensure_dirs(vdb_path)
    except portage.exception.PortageException:
        pass
    vdb_lock = None
    try:
        if os.access(vdb_path, os.W_OK):
            vartree.dbapi.lock()
            vdb_lock = True

        realsyslist = []
        sys_virt_map = {}
        for x in sets["system"].getAtoms():
            for atom in expand_new_virt(vartree.dbapi, x):
                if not atom.blocker:
                    realsyslist.append(atom)
                    if atom.cp != x.cp:
                        sys_virt_map[atom.cp] = x.cp

        syslist = []
        for x in realsyslist:
            mycp = x.cp
            # Since Gentoo stopped using old-style virtuals in
            # 2011, typically it's possible to avoid getvirtuals()
            # calls entirely. It will not be triggered here by
            # new-style virtuals since those are expanded to
            # non-virtual atoms above by expand_new_virt().
            if mycp.startswith("virtual/") and mycp in settings.getvirtuals():
                providers = []
                for provider in settings.getvirtuals()[mycp]:
                    if vartree.dbapi.match(provider):
                        providers.append(provider)
                if len(providers) == 1:
                    syslist.extend(providers)
            else:
                syslist.append(mycp)
        syslist = frozenset(syslist)

        if not unmerge_files:
            if unmerge_action in ["rage-clean", "unmerge"]:
                print()
                print(
                    bold("emerge %s" % unmerge_action)
                    + " can only be used with specific package names"
                )
                print()
                return 1, {}

            global_unmerge = 1

        # process all arguments and add all
        # valid db entries to candidate_catpkgs
        if global_unmerge:
            if not unmerge_files:
                candidate_catpkgs.extend(vartree.dbapi.cp_all())
        else:
            # we've got command-line arguments
            if not unmerge_files:
                print("\nNo packages to %s have been provided.\n" % unmerge_action)
                return 1, {}
            for x in unmerge_files:
                arg_parts = x.split("/")
                if x[0] not in [".", "/"] and arg_parts[-1][-7:] != ".ebuild":
                    # possible cat/pkg or dep; treat as such
                    candidate_catpkgs.append(x)
                elif unmerge_action in ["prune", "clean"]:
                    print(
                        "\n!!! Prune and clean do not accept individual"
                        + " ebuilds as arguments;\n    skipping.\n"
                    )
                    continue
                else:
                    # it appears that the user is specifying an installed
                    # ebuild and we're in "unmerge" mode, so it's ok.
                    if not os.path.exists(x):
                        print("\n!!! The path '" + x + "' doesn't exist.\n")
                        return 1, {}

                    absx = os.path.abspath(x)
                    sp_absx = absx.split("/")
                    if sp_absx[-1][-7:] == ".ebuild":
                        del sp_absx[-1]
                        absx = "/".join(sp_absx)

                    sp_absx_len = len(sp_absx)

                    vdb_path = os.path.join(settings["EROOT"], portage.VDB_PATH)

                    sp_vdb = vdb_path.split("/")
                    sp_vdb_len = len(sp_vdb)

                    if not os.path.exists(absx + "/CONTENTS"):
                        print("!!! Not a valid db dir: " + str(absx))
                        return 1, {}

                    if sp_absx_len <= sp_vdb_len:
                        # The Path is shorter... so it can't be inside the vdb.
                        print(sp_absx)
                        print(absx)
                        print(
                            "\n!!!", x, "cannot be inside " + vdb_path + "; aborting.\n"
                        )
                        return 1, {}

                    for idx in range(0, sp_vdb_len):
                        if idx >= sp_absx_len or sp_vdb[idx] != sp_absx[idx]:
                            print(sp_absx)
                            print(absx)
                            print(
                                "\n!!!",
                                x,
                                "is not inside " + vdb_path + "; aborting.\n",
                            )
                            return 1, {}

                    print("=" + "/".join(sp_absx[sp_vdb_len:]))
                    candidate_catpkgs.append("=" + "/".join(sp_absx[sp_vdb_len:]))

        newline = ""
        if not quiet:
            newline = "\n"
        if settings["ROOT"] != "/":
            writemsg_level(
                darkgreen(
                    newline
                    + ">>> Using system located in ROOT tree %s\n" % settings["ROOT"]
                )
            )

        if ("--pretend" in myopts or "--ask" in myopts) and not quiet:
            writemsg_level(
                darkgreen(
                    newline + ">>> These are the packages that would be unmerged:\n"
                )
            )

        # Preservation of order is required for --depclean and --prune so
        # that dependencies are respected. Use all_selected to eliminate
        # duplicate packages since the same package may be selected by
        # multiple atoms.
        pkgmap = []
        all_selected = set()
        for x in candidate_catpkgs:
            # cycle through all our candidate deps and determine
            # what will and will not get unmerged
            try:
                mymatch = vartree.dbapi.match(x)
            except portage.exception.AmbiguousPackageName as errpkgs:
                print(
                    '\n\n!!! The short ebuild name "'
                    + x
                    + '" is ambiguous.  Please specify'
                )
                print(
                    "!!! one of the following fully-qualified "
                    + "ebuild names instead:\n"
                )
                for i in errpkgs[0]:
                    print("    " + green(i))
                print()
                sys.exit(1)

            if not mymatch and x[0] not in "<>=~":
                mymatch = vartree.dep_match(x)
            if not mymatch:
                portage.writemsg(
                    "\n--- Couldn't find '%s' to %s.\n"
                    % (x.replace("null/", ""), unmerge_action),
                    noiselevel=-1,
                )
                continue

            pkgmap.append({"protected": set(), "selected": set(), "omitted": set()})
            mykey = len(pkgmap) - 1
            if unmerge_action in ["rage-clean", "unmerge"]:
                for y in mymatch:
                    if y not in all_selected:
                        pkgmap[mykey]["selected"].add(y)
                        all_selected.add(y)
            elif unmerge_action == "prune":
                if len(mymatch) == 1:
                    continue
                best_version = mymatch[0]
                best_slot = vartree.getslot(best_version)
                best_counter = vartree.dbapi.cpv_counter(best_version)
                for mypkg in mymatch[1:]:
                    myslot = vartree.getslot(mypkg)
                    mycounter = vartree.dbapi.cpv_counter(mypkg)
                    if (
                        myslot == best_slot and mycounter > best_counter
                    ) or mypkg == portage.best([mypkg, best_version]):
                        if myslot == best_slot:
                            if mycounter < best_counter:
                                # On slot collision, keep the one with the
                                # highest counter since it is the most
                                # recently installed.
                                continue
                        best_version = mypkg
                        best_slot = myslot
                        best_counter = mycounter
                pkgmap[mykey]["protected"].add(best_version)
                pkgmap[mykey]["selected"].update(
                    mypkg
                    for mypkg in mymatch
                    if mypkg != best_version and mypkg not in all_selected
                )
                all_selected.update(pkgmap[mykey]["selected"])
            else:
                # unmerge_action == "clean"
                slotmap = {}
                for mypkg in mymatch:
                    if unmerge_action == "clean":
                        myslot = vartree.getslot(mypkg)
                    else:
                        # since we're pruning, we don't care about slots
                        # and put all the pkgs in together
                        myslot = 0
                    if myslot not in slotmap:
                        slotmap[myslot] = {}
                    slotmap[myslot][vartree.dbapi.cpv_counter(mypkg)] = mypkg

                for mypkg in vartree.dbapi.cp_list(portage.cpv_getkey(mymatch[0])):
                    myslot = vartree.getslot(mypkg)
                    if myslot not in slotmap:
                        slotmap[myslot] = {}
                    slotmap[myslot][vartree.dbapi.cpv_counter(mypkg)] = mypkg

                for myslot in slotmap:
                    counterkeys = list(slotmap[myslot])
                    if not counterkeys:
                        continue
                    counterkeys.sort()
                    pkgmap[mykey]["protected"].add(slotmap[myslot][counterkeys[-1]])
                    del counterkeys[-1]

                    for counter in counterkeys[:]:
                        mypkg = slotmap[myslot][counter]
                        if mypkg not in mymatch:
                            counterkeys.remove(counter)
                            pkgmap[mykey]["protected"].add(slotmap[myslot][counter])

                    # be pretty and get them in order of merge:
                    for ckey in counterkeys:
                        mypkg = slotmap[myslot][ckey]
                        if mypkg not in all_selected:
                            pkgmap[mykey]["selected"].add(mypkg)
                            all_selected.add(mypkg)
                    # ok, now the last-merged package
                    # is protected, and the rest are selected
        numselected = len(all_selected)
        if global_unmerge and not numselected:
            portage.writemsg_stdout(
                "\n>>> No outdated packages were found on your system.\n"
            )
            return 1, {}

        if not numselected:
            portage.writemsg_stdout(
                "\n>>> No packages selected for removal by " + unmerge_action + "\n"
            )
            return 1, {}
    finally:
        if vdb_lock:
            vartree.dbapi.flush_cache()
            vartree.dbapi.unlock()

    # generate a list of package sets that are directly or indirectly listed in "selected",
    # as there is no persistent list of "installed" sets
    installed_sets = ["selected"]
    stop = False
    pos = 0
    while not stop:
        stop = True
        pos = len(installed_sets)
        for s in installed_sets[pos - 1 :]:
            if s not in sets:
                continue
            candidates = [
                x[len(SETPREFIX) :]
                for x in sets[s].getNonAtoms()
                if x.startswith(SETPREFIX)
            ]
            if candidates:
                stop = False
                installed_sets += candidates
    installed_sets = [
        x for x in installed_sets if x not in root_config.setconfig.active
    ]
    del stop, pos

    # we don't want to unmerge packages that are still listed in user-editable package sets
    # listed in "world" as they would be remerged on the next update of "world" or the
    # relevant package sets.
    unknown_sets = set()
    for cp in range(len(pkgmap)):
        for cpv in pkgmap[cp]["selected"].copy():
            try:
                pkg = _pkg(cpv)
            except KeyError:
                # It could have been uninstalled
                # by a concurrent process.
                continue

            if unmerge_action != "clean" and root_config.root == "/":
                skip_pkg = False
                if portage.match_from_list(portage.const.PORTAGE_PACKAGE_ATOM, [pkg]):
                    msg = (
                        "Not unmerging package %s "
                        "since there is no valid reason for Portage to "
                        "%s itself."
                    ) % (pkg.cpv, unmerge_action)
                    skip_pkg = True
                elif vartree.dbapi._dblink(cpv).isowner(portage._python_interpreter):
                    msg = (
                        "Not unmerging package %s since there is no valid "
                        "reason for Portage to %s currently used Python "
                        "interpreter."
                    ) % (pkg.cpv, unmerge_action)
                    skip_pkg = True
                if skip_pkg:
                    for line in textwrap.wrap(msg, 75):
                        out.eerror(line)
                    # adjust pkgmap so the display output is correct
                    pkgmap[cp]["selected"].remove(cpv)
                    all_selected.remove(cpv)
                    pkgmap[cp]["protected"].add(cpv)
                    continue

            parents = []
            for s in installed_sets:
                # skip sets that the user requested to unmerge, and skip world
                # user-selected set, since the package will be removed from
                # that set later on.
                if s in root_config.setconfig.active or s == "selected":
                    continue

                if s not in sets:
                    if s in unknown_sets:
                        continue
                    unknown_sets.add(s)
                    out = portage.output.EOutput()
                    out.eerror(
                        ("Unknown set '@%s' in %s%s")
                        % (
                            s,
                            root_config.settings["EROOT"],
                            portage.const.WORLD_SETS_FILE,
                        )
                    )
                    continue

                # only check instances of EditablePackageSet as other classes are generally used for
                # special purposes and can be ignored here (and are usually generated dynamically, so the
                # user can't do much about them anyway)
                if isinstance(sets[s], EditablePackageSet):

                    # This is derived from a snippet of code in the
                    # depgraph._iter_atoms_for_pkg() method.
                    for atom in sets[s].iterAtomsForPackage(pkg):
                        inst_matches = vartree.dbapi.match(atom)
                        inst_matches.reverse()  # descending order
                        higher_slot = None
                        for inst_cpv in inst_matches:
                            try:
                                inst_pkg = _pkg(inst_cpv)
                            except KeyError:
                                # It could have been uninstalled
                                # by a concurrent process.
                                continue

                            if inst_pkg.cp != atom.cp:
                                continue
                            if pkg >= inst_pkg:
                                # This is descending order, and we're not
                                # interested in any versions <= pkg given.
                                break
                            if pkg.slot_atom != inst_pkg.slot_atom:
                                higher_slot = inst_pkg
                                break
                        if higher_slot is None:
                            parents.append(s)
                            break
            if parents:
                print(colorize("WARN", "Package %s is going to be unmerged," % cpv))
                print(
                    colorize("WARN", "but still listed in the following package sets:")
                )
                print("    %s\n" % ", ".join(parents))

    del installed_sets

    numselected = len(all_selected)
    if not numselected:
        writemsg_level(
            "\n>>> No packages selected for removal by " + unmerge_action + "\n"
        )
        return 1, {}

    # Unmerge order only matters in some cases
    if not ordered:
        unordered = {}
        for d in pkgmap:
            selected = d["selected"]
            if not selected:
                continue
            cp = portage.cpv_getkey(next(iter(selected)))
            cp_dict = unordered.get(cp)
            if cp_dict is None:
                cp_dict = {}
                unordered[cp] = cp_dict
                for k in d:
                    cp_dict[k] = set()
            for k, v in d.items():
                cp_dict[k].update(v)
        pkgmap = [unordered[cp] for cp in sorted(unordered)]

    # Sort each set of selected packages
    if ordered:
        for pkg in pkgmap:
            pkg["selected"] = sorted(pkg["selected"], key=cpv_sort_key())

    for x in range(len(pkgmap)):
        selected = pkgmap[x]["selected"]
        if not selected:
            continue
        for mytype, mylist in pkgmap[x].items():
            if mytype == "selected":
                continue
            mylist.difference_update(all_selected)
        cp = portage.cpv_getkey(next(iter(selected)))
        for y in vartree.dep_match(cp):
            if (
                y not in pkgmap[x]["omitted"]
                and y not in pkgmap[x]["selected"]
                and y not in pkgmap[x]["protected"]
                and y not in all_selected
            ):
                pkgmap[x]["omitted"].add(y)
        if global_unmerge and not pkgmap[x]["selected"]:
            # avoid cluttering the preview printout with stuff that isn't getting unmerged
            continue
        if not (pkgmap[x]["protected"] or pkgmap[x]["omitted"]) and cp in syslist:
            virt_cp = sys_virt_map.get(cp)
            if virt_cp is None:
                cp_info = "'{}'".format(cp)
            else:
                cp_info = "'{}' ({})".format(cp, virt_cp)
            writemsg_level(
                colorize(
                    "BAD",
                    "\n\n!!! " + "{} is part of your system profile.\n".format(cp_info),
                ),
                level=logging.WARNING,
                noiselevel=-1,
            )
            writemsg_level(
                colorize(
                    "WARN", "!!! Unmerging it may " + "be damaging to your system.\n\n"
                ),
                level=logging.WARNING,
                noiselevel=-1,
            )
        if not quiet:
            writemsg_level("\n {}\n".format(bold(cp)), noiselevel=-1)
        else:
            writemsg_level(bold(cp) + ": ", noiselevel=-1)
        for mytype in ["selected", "protected", "omitted"]:
            if not quiet:
                writemsg_level((mytype + ": ").rjust(14), noiselevel=-1)
            if pkgmap[x][mytype]:
                sorted_pkgs = []
                for mypkg in pkgmap[x][mytype]:
                    try:
                        sorted_pkgs.append(mypkg.cpv)
                    except AttributeError:
                        sorted_pkgs.append(_pkg_str(mypkg))
                sorted_pkgs.sort(key=cpv_sort_key())
                for mypkg in sorted_pkgs:
                    if mytype == "selected":
                        writemsg_level(
                            colorize("UNMERGE_WARN", mypkg.version + " "), noiselevel=-1
                        )
                    else:
                        writemsg_level(
                            colorize("GOOD", mypkg.version + " "), noiselevel=-1
                        )
            else:
                writemsg_level("none ", noiselevel=-1)
            if not quiet:
                writemsg_level("\n", noiselevel=-1)
        if quiet:
            writemsg_level("\n", noiselevel=-1)

    writemsg_level(
        "\nAll selected packages: %s\n" % " ".join("=%s" % x for x in all_selected),
        noiselevel=-1,
    )

    writemsg_level(
        "\n>>> "
        + colorize("UNMERGE_WARN", "'Selected'")
        + " packages are slated for removal.\n"
    )
    writemsg_level(
        ">>> "
        + colorize("GOOD", "'Protected'")
        + " and "
        + colorize("GOOD", "'omitted'")
        + " packages will not be removed.\n\n"
    )

    return os.EX_OK, pkgmap


def unmerge(
    root_config,
    myopts,
    unmerge_action,
    unmerge_files,
    ldpath_mtimes,
    autoclean=0,
    clean_world=1,
    clean_delay=1,
    ordered=0,
    raise_on_error=0,
    scheduler=None,
    writemsg_level=portage.util.writemsg_level,
):
    """
    Returns os.EX_OK if no errors occur, 1 if an error occurs, and
    130 if interrupted due to a 'no' answer for --ask.
    """

    if clean_world:
        clean_world = myopts.get("--deselect") != "n"

    rval, pkgmap = _unmerge_display(
        root_config,
        myopts,
        unmerge_action,
        unmerge_files,
        clean_delay=clean_delay,
        ordered=ordered,
        writemsg_level=writemsg_level,
    )

    if rval != os.EX_OK:
        return rval

    enter_invalid = "--ask-enter-invalid" in myopts
    vartree = root_config.trees["vartree"]
    sets = root_config.sets
    settings = root_config.settings
    mysettings = portage.config(clone=settings)
    xterm_titles = "notitles" not in settings.features

    if "--pretend" in myopts:
        # we're done... return
        return os.EX_OK
    if "--ask" in myopts:
        uq = UserQuery(myopts)
        if uq.query("Would you like to unmerge these packages?", enter_invalid) == "No":
            # enter pretend mode for correct formatting of results
            myopts["--pretend"] = True
            print()
            print("Quitting.")
            print()
            return 128 + signal.SIGINT

    if not vartree.dbapi.writable:
        writemsg_level(
            "!!! %s\n" % _("Read-only file system: %s") % vartree.dbapi._dbroot,
            level=logging.ERROR,
            noiselevel=-1,
        )
        return 1

    # the real unmerging begins, after a short delay unless we're raging....
    if not unmerge_action == "rage-clean" and clean_delay and not autoclean:
        countdown(int(settings["CLEAN_DELAY"]), ">>> Unmerging")

    all_selected = set()
    all_selected.update(*[x["selected"] for x in pkgmap])

    # Set counter variables
    curval = 1
    maxval = len(all_selected)

    for x in range(len(pkgmap)):
        for y in pkgmap[x]["selected"]:
            emergelog(xterm_titles, "=== Unmerging... (" + y + ")")
            message = ">>> Unmerging ({} of {}) {}...\n".format(
                colorize("MERGE_LIST_PROGRESS", str(curval)),
                colorize("MERGE_LIST_PROGRESS", str(maxval)),
                y,
            )
            writemsg_level(message, noiselevel=-1)
            curval += 1

            mysplit = y.split("/")
            # unmerge...
            retval = portage.unmerge(
                mysplit[0],
                mysplit[1],
                settings=mysettings,
                vartree=vartree,
                ldpath_mtimes=ldpath_mtimes,
                scheduler=scheduler,
            )

            if retval != os.EX_OK:
                emergelog(xterm_titles, " !!! unmerge FAILURE: " + y)
                if raise_on_error:
                    raise UninstallFailure(retval)
                sys.exit(retval)
            else:
                if (
                    clean_world
                    and hasattr(sets["selected"], "cleanPackage")
                    and hasattr(sets["selected"], "lock")
                ):
                    sets["selected"].lock()
                    if hasattr(sets["selected"], "load"):
                        sets["selected"].load()
                    sets["selected"].cleanPackage(vartree.dbapi, y)
                    sets["selected"].unlock()
                emergelog(xterm_titles, " >>> unmerge success: " + y)

    if (
        clean_world
        and hasattr(sets["selected"], "remove")
        and hasattr(sets["selected"], "lock")
    ):
        sets["selected"].lock()
        # load is called inside remove()
        for s in root_config.setconfig.active:
            sets["selected"].remove(SETPREFIX + s)
        sets["selected"].unlock()

    return os.EX_OK



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import array
import errno
import os

from portage.util.futures import asyncio
from _emerge.AsynchronousTask import AsynchronousTask


class AbstractPollTask(AsynchronousTask):

    __slots__ = ("_registered",)

    _bufsize = 4096

    def _read_array(self, f):
        """
        NOTE: array.fromfile() is used here only for testing purposes,
        because it has bugs in all known versions of Python (including
        Python 2.7 and Python 3.2). See PipeReaderArrayTestCase.

        A benchmark that copies bytes from /dev/zero to /dev/null shows
        that arrays give a 15% performance improvement for Python 2.7.14.
        However, arrays significantly *decrease* performance for Python 3.
        """
        buf = array.array("B")
        try:
            buf.fromfile(f, self._bufsize)
        except EOFError:
            pass
        except TypeError:
            # Python 3.2:
            # TypeError: read() didn't return bytes
            pass
        except OSError as e:
            # EIO happens with pty on Linux after the
            # slave end of the pty has been closed.
            if e.errno == errno.EIO:
                # EOF: return empty string of bytes
                pass
            elif e.errno == errno.EAGAIN:
                # EAGAIN: return None
                buf = None
            else:
                raise

        if buf is not None:
            try:
                # Python >=3.2
                buf = buf.tobytes()
            except AttributeError:
                buf = buf.tostring()

        return buf

    def _read_buf(self, fd):
        """
        Read self._bufsize into a string of bytes, handling EAGAIN and
        EIO. This will only call os.read() once, so the caller should
        call this method in a loop until either None or an empty string
        of bytes is returned. An empty string of bytes indicates EOF.
        None indicates EAGAIN.

        NOTE: os.read() will be called regardless of the event flags,
                since otherwise data may be lost (see bug #531724).

        @param fd: file descriptor (non-blocking mode required)
        @type fd: int
        @rtype: bytes or None
        @return: A string of bytes, or None
        """
        # NOTE: array.fromfile() is no longer used here because it has
        # bugs in all known versions of Python (including Python 2.7
        # and Python 3.2).
        buf = None
        try:
            buf = os.read(fd, self._bufsize)
        except OSError as e:
            # EIO happens with pty on Linux after the
            # slave end of the pty has been closed.
            if e.errno == errno.EIO:
                # EOF: return empty string of bytes
                buf = b""
            elif e.errno == errno.EAGAIN:
                # EAGAIN: return None
                buf = None
            else:
                raise

        return buf

    def _async_wait(self):
        self._unregister()
        super()._async_wait()

    def _unregister(self):
        self._registered = False

    def _wait_loop(self, timeout=None):
        loop = self.scheduler
        tasks = [self.async_wait()]
        if timeout is not None:
            tasks.append(
                asyncio.ensure_future(asyncio.sleep(timeout, loop=loop), loop=loop)
            )
        try:
            loop.run_until_complete(
                asyncio.ensure_future(
                    asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED, loop=loop),
                    loop=loop,
                )
            )
        finally:
            for task in tasks:
                task.cancel()



# Copyright 2020-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import subprocess

from _emerge.AsynchronousTask import AsynchronousTask

from portage import os
from portage.util import shlex_split
from portage.util._async.PipeLogger import PipeLogger
from portage.util._async.PopenProcess import PopenProcess
from portage.util.futures import asyncio


class BuildLogger(AsynchronousTask):
    """
    Write to a log file, with compression support provided by PipeLogger.
    If the log_filter_file parameter is specified, then it is interpreted
    as a command to execute which filters log output (see the
    PORTAGE_LOG_FILTER_FILE_CMD variable in make.conf(5)). The stdin property
    provides access to a writable binary file stream (refers to a pipe)
    that log content should be written to (usually redirected from
    subprocess stdout and stderr streams).
    """

    __slots__ = (
        "env",
        "log_path",
        "log_filter_file",
        "_main_task",
        "_main_task_cancel",
        "_stdin",
    )

    @property
    def stdin(self):
        return self._stdin

    def _start(self):
        filter_proc = None
        log_input = None
        if self.log_path is not None:
            log_filter_file = self.log_filter_file
            if log_filter_file is not None:
                split_value = shlex_split(log_filter_file)
                log_filter_file = split_value if split_value else None
            if log_filter_file:
                filter_input, stdin = os.pipe()
                log_input, filter_output = os.pipe()
                try:
                    filter_proc = PopenProcess(
                        proc=subprocess.Popen(
                            log_filter_file,
                            env=self.env,
                            stdin=filter_input,
                            stdout=filter_output,
                            stderr=filter_output,
                        ),
                        scheduler=self.scheduler,
                    )
                    filter_proc.start()
                except OSError:
                    # Maybe the command is missing or broken somehow...
                    os.close(filter_input)
                    os.close(stdin)
                    os.close(log_input)
                    os.close(filter_output)
                else:
                    self._stdin = os.fdopen(stdin, "wb", 0)
                    os.close(filter_input)
                    os.close(filter_output)

        if self._stdin is None:
            # Since log_filter_file is unspecified or refers to a file
            # that is missing or broken somehow, create a pipe that
            # logs directly to pipe_logger.
            log_input, stdin = os.pipe()
            self._stdin = os.fdopen(stdin, "wb", 0)

        # Set background=True so that pipe_logger does not log to stdout.
        pipe_logger = PipeLogger(
            background=True,
            scheduler=self.scheduler,
            input_fd=log_input,
            log_file_path=self.log_path,
        )
        pipe_logger.start()

        self._main_task_cancel = functools.partial(
            self._main_cancel, filter_proc, pipe_logger
        )
        self._main_task = asyncio.ensure_future(
            self._main(filter_proc, pipe_logger), loop=self.scheduler
        )
        self._main_task.add_done_callback(self._main_exit)

    async def _main(self, filter_proc, pipe_logger):
        try:
            if pipe_logger.poll() is None:
                await pipe_logger.async_wait()
            if filter_proc is not None and filter_proc.poll() is None:
                await filter_proc.async_wait()
        except asyncio.CancelledError:
            self._main_cancel(filter_proc, pipe_logger)
            raise

    def _main_cancel(self, filter_proc, pipe_logger):
        if pipe_logger.poll() is None:
            pipe_logger.cancel()
        if filter_proc is not None and filter_proc.poll() is None:
            filter_proc.cancel()

    def _cancel(self):
        if self._main_task is not None:
            if not self._main_task.done():
                if self._main_task_cancel is not None:
                    self._main_task_cancel()
                    self._main_task_cancel = None
                self._main_task.cancel()
        if self._stdin is not None and not self._stdin.closed:
            self._stdin.close()

    def _main_exit(self, main_task):
        self._main_task = None
        self._main_task_cancel = None
        try:
            main_task.result()
        except asyncio.CancelledError:
            self.cancel()
            self._was_cancelled()
        self.returncode = self.returncode or 0
        self._async_wait()



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class AutounmaskBinpkgUseTestCase(TestCase):
    def testAutounmaskBinpkgUse(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/B[foo]",
                "RDEPEND": "dev-libs/B[foo]",
            },
            "dev-libs/B-1": {
                "EAPI": "6",
                "IUSE": "foo",
            },
        }
        binpkgs = {
            "dev-libs/A-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/B[foo]",
                "RDEPEND": "dev-libs/B[foo]",
            },
            "dev-libs/B-1": {
                "EAPI": "6",
                "IUSE": "foo",
                "USE": "foo",
            },
        }
        installed = {}

        test_cases = (
            # Bug 619626: Test for unnecessary rebuild due
            # to rejection of binary packages that would
            # be acceptable after appplication of autounmask
            # USE changes.
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                all_permutations=True,
                success=True,
                options={
                    "--usepkg": True,
                },
                mergelist=[
                    "[binary]dev-libs/B-1",
                    "[binary]dev-libs/A-1",
                ],
                use_changes={"dev-libs/B-1": {"foo": True}},
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()



# Copyright 2010-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import signal
import tempfile

try:
    import dummy_threading
except ImportError:
    dummy_threading = None

from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from _emerge.AsynchronousLock import AsynchronousLock


class AsynchronousLockTestCase(TestCase):
    def _testAsynchronousLock(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            for force_async in (True, False):
                for force_dummy in (
                    (False,) if dummy_threading is None else (True, False)
                ):
                    async_lock = AsynchronousLock(
                        path=path,
                        scheduler=scheduler,
                        _force_async=force_async,
                        _force_thread=True,
                        _force_dummy=force_dummy,
                    )
                    async_lock.start()
                    self.assertEqual(async_lock.wait(), os.EX_OK)
                    self.assertEqual(async_lock.returncode, os.EX_OK)
                    scheduler.run_until_complete(async_lock.async_unlock())

                async_lock = AsynchronousLock(
                    path=path,
                    scheduler=scheduler,
                    _force_async=force_async,
                    _force_process=True,
                )
                async_lock.start()
                self.assertEqual(async_lock.wait(), os.EX_OK)
                self.assertEqual(async_lock.returncode, os.EX_OK)
                scheduler.run_until_complete(async_lock.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    def testAsynchronousLock(self):
        self._testAsynchronousLock()

    def testAsynchronousLockHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLock()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def _testAsynchronousLockWait(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = AsynchronousLock(path=path, scheduler=scheduler)
            lock1.start()
            self.assertEqual(lock1.wait(), os.EX_OK)
            self.assertEqual(lock1.returncode, os.EX_OK)

            # lock2 requires _force_async=True since the portage.locks
            # module is not designed to work as intended here if the
            # same process tries to lock the same file more than
            # one time concurrently.
            lock2 = AsynchronousLock(
                path=path, scheduler=scheduler, _force_async=True, _force_process=True
            )
            lock2.start()
            # lock2 should be waiting for lock1 to release
            self.assertEqual(lock2.poll(), None)
            self.assertEqual(lock2.returncode, None)

            scheduler.run_until_complete(lock1.async_unlock())
            self.assertEqual(lock2.wait(), os.EX_OK)
            self.assertEqual(lock2.returncode, os.EX_OK)
            scheduler.run_until_complete(lock2.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    def testAsynchronousLockWait(self):
        self._testAsynchronousLockWait()

    def testAsynchronousLockWaitHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLockWait()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def _testAsynchronousLockWaitCancel(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = AsynchronousLock(path=path, scheduler=scheduler)
            lock1.start()
            self.assertEqual(lock1.wait(), os.EX_OK)
            self.assertEqual(lock1.returncode, os.EX_OK)
            lock2 = AsynchronousLock(
                path=path, scheduler=scheduler, _force_async=True, _force_process=True
            )
            lock2.start()
            # lock2 should be waiting for lock1 to release
            self.assertEqual(lock2.poll(), None)
            self.assertEqual(lock2.returncode, None)

            # Cancel lock2 and then check wait() and returncode results.
            lock2.cancel()
            self.assertEqual(lock2.wait() == os.EX_OK, False)
            self.assertEqual(lock2.returncode == os.EX_OK, False)
            self.assertEqual(lock2.returncode is None, False)
            scheduler.run_until_complete(lock1.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    def testAsynchronousLockWaitCancel(self):
        self._testAsynchronousLockWaitCancel()

    def testAsynchronousLockWaitCancelHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLockWaitCancel()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def _testAsynchronousLockWaitKill(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = AsynchronousLock(path=path, scheduler=scheduler)
            lock1.start()
            self.assertEqual(lock1.wait(), os.EX_OK)
            self.assertEqual(lock1.returncode, os.EX_OK)
            lock2 = AsynchronousLock(
                path=path, scheduler=scheduler, _force_async=True, _force_process=True
            )
            lock2.start()
            # lock2 should be waiting for lock1 to release
            self.assertEqual(lock2.poll(), None)
            self.assertEqual(lock2.returncode, None)

            # Kill lock2's process and then check wait() and
            # returncode results. This is intended to simulate
            # a SIGINT sent via the controlling tty.
            self.assertEqual(lock2._imp is not None, True)
            self.assertEqual(lock2._imp._proc is not None, True)
            self.assertEqual(lock2._imp._proc.pid is not None, True)
            lock2._imp._kill_test = True
            os.kill(lock2._imp._proc.pid, signal.SIGTERM)
            self.assertEqual(lock2.wait() == os.EX_OK, False)
            self.assertEqual(lock2.returncode == os.EX_OK, False)
            self.assertEqual(lock2.returncode is None, False)
            scheduler.run_until_complete(lock1.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    def testAsynchronousLockWaitKill(self):
        self._testAsynchronousLockWaitKill()

    def testAsynchronousLockWaitKillHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLockWaitKill()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state



# Copyright 1998-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["close_portdbapi_caches", "FetchlistDict", "portagetree", "portdbapi"]

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.checksum",
    "portage.data:portage_gid,secpass",
    "portage.dbapi.dep_expand:dep_expand",
    "portage.dep:Atom,dep_getkey,match_from_list,use_reduce,_match_slot",
    "portage.package.ebuild.doebuild:doebuild",
    "portage.package.ebuild.fetch:get_mirror_url,_download_suffix",
    "portage.util:ensure_dirs,shlex_split,writemsg,writemsg_level",
    "portage.util.listdir:listdir",
    "portage.versions:best,catsplit,catpkgsplit,_pkgsplit@pkgsplit,ver_regexp,_pkg_str",
)

from portage.cache import volatile
from portage.cache.cache_errors import CacheError
from portage.cache.mappings import Mapping
from portage.dbapi import dbapi
from portage.exception import (
    PortageException,
    PortageKeyError,
    FileNotFound,
    InvalidAtom,
    InvalidData,
    InvalidDependString,
    InvalidPackageName,
)
from portage.localization import _

from portage import eclass_cache, eapi_is_supported, _eapi_is_deprecated
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage.util.futures import asyncio
from portage.util.futures.iter_completed import iter_gather
from _emerge.EbuildMetadataPhase import EbuildMetadataPhase

import os as _os
import traceback
import warnings
import errno
import functools

import collections
from collections import OrderedDict
from urllib.parse import urlparse


def close_portdbapi_caches():
    # The python interpreter does _not_ guarantee that destructors are
    # called for objects that remain when the interpreter exits, so we
    # use an atexit hook to call destructors for any global portdbapi
    # instances that may have been constructed.
    try:
        portage._legacy_globals_constructed
    except AttributeError:
        pass
    else:
        if "db" in portage._legacy_globals_constructed:
            try:
                db = portage.db
            except AttributeError:
                pass
            else:
                if isinstance(db, dict):
                    for x in db.values():
                        try:
                            if "porttree" in x.lazy_items:
                                continue
                        except (AttributeError, TypeError):
                            continue
                        try:
                            x = x.pop("porttree").dbapi
                        except (AttributeError, KeyError):
                            continue
                        if not isinstance(x, portdbapi):
                            continue
                        x.close_caches()


portage.process.atexit_register(close_portdbapi_caches)

# It used to be necessary for API consumers to remove portdbapi instances
# from portdbapi_instances, in order to avoid having accumulated instances
# consume memory. Now, portdbapi_instances is just an empty dummy list, so
# for backward compatibility, ignore ValueError for removal on non-existent
# items.


class _dummy_list(list):
    def remove(self, item):
        # TODO: Trigger a DeprecationWarning here, after stable portage
        # has dummy portdbapi_instances.
        try:
            list.remove(self, item)
        except ValueError:
            pass


class _better_cache:

    """
    The purpose of better_cache is to locate catpkgs in repositories using ``os.listdir()`` as much as possible, which
    is less expensive IO-wise than exhaustively doing a stat on each repo for a particular catpkg. better_cache stores a
    list of repos in which particular catpkgs appear. Various dbapi methods use better_cache to locate repositories of
    interest related to particular catpkg rather than performing an exhaustive scan of all repos/overlays.

    Better_cache.items data may look like this::

      { "sys-apps/portage" : [ repo1, repo2 ] }

    Without better_cache, Portage will get slower and slower (due to excessive IO) as more overlays are added.

    Also note that it is OK if this cache has some 'false positive' catpkgs in it. We use it to search for specific
    catpkgs listed in ebuilds. The likelihood of a false positive catpkg in our cache causing a problem is extremely
    low, because the user of our cache is passing us a catpkg that came from somewhere and has already undergone some
    validation, and even then will further interrogate the short-list of repos we return to gather more information
    on the catpkg.

    Thus, the code below is optimized for speed rather than painstaking correctness. I have added a note to
    ``dbapi.getRepositories()`` to ensure that developers are aware of this just in case.

    The better_cache has been redesigned to perform on-demand scans -- it will only scan a category at a time, as
    needed. This should further optimize IO performance by not scanning category directories that are not needed by
    Portage.
    """

    def __init__(self, repositories):
        self._items = collections.defaultdict(list)
        self._scanned_cats = set()

        # ordered list of all portree locations we'll scan:
        self._repo_list = [
            repo for repo in reversed(list(repositories)) if repo.location is not None
        ]

    def __getitem__(self, catpkg):
        result = self._items.get(catpkg)
        if result is not None:
            return result

        cat, pkg = catsplit(catpkg)
        if cat not in self._scanned_cats:
            self._scan_cat(cat)
        return self._items[catpkg]

    def _scan_cat(self, cat):
        for repo in self._repo_list:
            cat_dir = repo.location + "/" + cat
            try:
                pkg_list = os.listdir(cat_dir)
            except OSError as e:
                if e.errno not in (errno.ENOTDIR, errno.ENOENT, errno.ESTALE):
                    raise
                continue
            for p in pkg_list:
                try:
                    atom = Atom("{}/{}".format(cat, p))
                except InvalidAtom:
                    continue
                if atom != atom.cp:
                    continue
                self._items[atom.cp].append(repo)
        self._scanned_cats.add(cat)


class portdbapi(dbapi):
    """this tree will scan a portage directory located at root (passed to init)"""

    portdbapi_instances = _dummy_list()
    _use_mutable = True

    @property
    def _categories(self):
        return self.settings.categories

    @property
    def porttree_root(self):
        warnings.warn(
            "portage.dbapi.porttree.portdbapi.porttree_root is deprecated in favor of portage.repository.config.RepoConfig.location "
            "(available as repositories[repo_name].location attribute of instances of portage.dbapi.porttree.portdbapi class)",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.settings.repositories.mainRepoLocation()

    @property
    def eclassdb(self):
        warnings.warn(
            "portage.dbapi.porttree.portdbapi.eclassdb is deprecated in favor of portage.repository.config.RepoConfig.eclass_db "
            "(available as repositories[repo_name].eclass_db attribute of instances of portage.dbapi.porttree.portdbapi class)",
            DeprecationWarning,
            stacklevel=2,
        )
        main_repo = self.repositories.mainRepo()
        if main_repo is None:
            return None
        return main_repo.eclass_db

    def __init__(self, _unused_param=DeprecationWarning, mysettings=None):
        """
        @param _unused_param: deprecated, use mysettings['PORTDIR'] instead
        @type _unused_param: None
        @param mysettings: an immutable config instance
        @type mysettings: portage.config
        """

        from portage import config

        if mysettings:
            self.settings = mysettings
        else:
            from portage import settings

            self.settings = config(clone=settings)

        if _unused_param is not DeprecationWarning:
            warnings.warn(
                "The first parameter of the "
                + "portage.dbapi.porttree.portdbapi"
                + " constructor is unused since portage-2.1.8. "
                + "mysettings['PORTDIR'] is used instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.repositories = self.settings.repositories
        self.treemap = self.repositories.treemap

        # This is strictly for use in aux_get() doebuild calls when metadata
        # is generated by the depend phase.  It's safest to use a clone for
        # this purpose because doebuild makes many changes to the config
        # instance that is passed in.
        self.doebuild_settings = config(clone=self.settings)
        self.depcachedir = os.path.realpath(self.settings.depcachedir)

        if os.environ.get("SANDBOX_ON") == "1":
            # Make api consumers exempt from sandbox violations
            # when doing metadata cache updates.
            sandbox_write = os.environ.get("SANDBOX_WRITE", "").split(":")
            if self.depcachedir not in sandbox_write:
                sandbox_write.append(self.depcachedir)
                os.environ["SANDBOX_WRITE"] = ":".join(filter(None, sandbox_write))

        self.porttrees = list(self.settings.repositories.repoLocationList())

        # This is used as sanity check for aux_get(). If there is no
        # root eclass dir, we assume that PORTDIR is invalid or
        # missing. This check allows aux_get() to detect a missing
        # repository and return early by raising a KeyError.
        self._have_root_eclass_dir = os.path.isdir(
            os.path.join(self.settings.repositories.mainRepoLocation(), "eclass")
        )

        # if the portdbapi is "frozen", then we assume that we can cache everything (that no updates to it are happening)
        self.xcache = {}
        self.frozen = 0

        # Keep a list of repo names, sorted by priority (highest priority first).
        self._ordered_repo_name_list = tuple(reversed(self.repositories.prepos_order))

        self.auxdbmodule = self.settings.load_best_module("portdbapi.auxdbmodule")
        self.auxdb = {}
        self._pregen_auxdb = {}
        # If the current user doesn't have depcachedir write permission,
        # then the depcachedir cache is kept here read-only access.
        self._ro_auxdb = {}
        self._init_cache_dirs()
        try:
            depcachedir_st = os.stat(self.depcachedir)
            depcachedir_w_ok = os.access(self.depcachedir, os.W_OK)
        except OSError:
            depcachedir_st = None
            depcachedir_w_ok = False

        cache_kwargs = {}

        depcachedir_unshared = False
        if (
            portage.data.secpass < 1
            and depcachedir_w_ok
            and depcachedir_st is not None
            and os.getuid() == depcachedir_st.st_uid
            and os.getgid() == depcachedir_st.st_gid
        ):
            # If this user owns depcachedir and is not in the
            # portage group, then don't bother to set permissions
            # on cache entries. This makes it possible to run
            # egencache without any need to be a member of the
            # portage group.
            depcachedir_unshared = True
        else:
            cache_kwargs.update({"gid": portage_gid, "perms": 0o664})

        # If secpass < 1, we don't want to write to the cache
        # since then we won't be able to apply group permissions
        # to the cache entries/directories.
        if (secpass < 1 and not depcachedir_unshared) or not depcachedir_w_ok:
            for x in self.porttrees:
                self.auxdb[x] = volatile.database(
                    self.depcachedir, x, self._known_keys, **cache_kwargs
                )
                try:
                    self._ro_auxdb[x] = self.auxdbmodule(
                        self.depcachedir,
                        x,
                        self._known_keys,
                        readonly=True,
                        **cache_kwargs
                    )
                except CacheError:
                    pass
        else:
            for x in self.porttrees:
                if x in self.auxdb:
                    continue
                # location, label, auxdbkeys
                self.auxdb[x] = self.auxdbmodule(
                    self.depcachedir, x, self._known_keys, **cache_kwargs
                )
        if "metadata-transfer" not in self.settings.features:
            for x in self.porttrees:
                if x in self._pregen_auxdb:
                    continue
                cache = self._create_pregen_cache(x)
                if cache is not None:
                    self._pregen_auxdb[x] = cache
        # Selectively cache metadata in order to optimize dep matching.
        self._aux_cache_keys = {
            "BDEPEND",
            "DEPEND",
            "EAPI",
            "IDEPEND",
            "INHERITED",
            "IUSE",
            "KEYWORDS",
            "LICENSE",
            "PDEPEND",
            "PROPERTIES",
            "RDEPEND",
            "repository",
            "RESTRICT",
            "SLOT",
            "DEFINED_PHASES",
            "REQUIRED_USE",
        }

        self._aux_cache = {}
        self._better_cache = None
        self._broken_ebuilds = set()

    def _set_porttrees(self, porttrees):
        """
        Consumers, such as emirrordist, may modify the porttrees attribute in
        order to modify the effective set of repositories for all portdbapi
        operations.

        @param porttrees: list of repo locations, in ascending order by
                repo priority
        @type porttrees: list
        """
        self._porttrees = tuple(porttrees)
        self._porttrees_repos = portage.OrderedDict(
            (repo.name, repo)
            for repo in (
                self.repositories.get_repo_for_location(location)
                for location in self._porttrees
            )
        )

    def _get_porttrees(self):
        return self._porttrees

    porttrees = property(_get_porttrees, _set_porttrees)

    @property
    def _event_loop(self):
        return asyncio._safe_loop()

    def _create_pregen_cache(self, tree):
        conf = self.repositories.get_repo_for_location(tree)
        cache = conf.get_pregenerated_cache(self._known_keys, readonly=True)
        if cache is not None:
            try:
                cache.ec = self.repositories.get_repo_for_location(tree).eclass_db
            except AttributeError:
                pass

            if not cache.complete_eclass_entries:
                warnings.warn(
                    (
                        "Repository '%s' used deprecated 'pms' cache format. "
                        "Please migrate to 'md5-dict' format."
                    )
                    % (conf.name,),
                    DeprecationWarning,
                )

        return cache

    def _init_cache_dirs(self):
        """Create /var/cache/edb/dep and adjust permissions for the portage
        group."""

        dirmode = 0o2070
        modemask = 0o2

        try:
            ensure_dirs(self.depcachedir, gid=portage_gid, mode=dirmode, mask=modemask)
        except PortageException:
            pass

    def close_caches(self):
        if not hasattr(self, "auxdb"):
            # unhandled exception thrown from constructor
            return
        for x in self.auxdb:
            self.auxdb[x].sync()
        self.auxdb.clear()

    def flush_cache(self):
        for x in self.auxdb.values():
            x.sync()

    def findLicensePath(self, license_name):
        for x in reversed(self.porttrees):
            license_path = os.path.join(x, "licenses", license_name)
            if os.access(license_path, os.R_OK):
                return license_path
        return None

    def findname(self, mycpv, mytree=None, myrepo=None):
        return self.findname2(mycpv, mytree, myrepo)[0]

    def getRepositoryPath(self, repository_id):
        """
        This function is required for GLEP 42 compliance; given a valid repository ID
        it must return a path to the repository
        TreeMap = { id:path }
        """
        return self.treemap.get(repository_id)

    def getRepositoryName(self, canonical_repo_path):
        """
        This is the inverse of getRepositoryPath().
        @param canonical_repo_path: the canonical path of a repository, as
                resolved by os.path.realpath()
        @type canonical_repo_path: String
        @return: The repo_name for the corresponding repository, or None
                if the path does not correspond a known repository
        @rtype: String or None
        """
        try:
            return self.repositories.get_name_for_location(canonical_repo_path)
        except KeyError:
            return None

    def getRepositories(self, catpkg=None):
        """
        With catpkg=None, this will return a complete list of repositories in this dbapi. With catpkg set to a value,
        this method will return a short-list of repositories that contain this catpkg. Use this second approach if
        possible, to avoid exhaustively searching all repos for a particular catpkg. It's faster for this method to
        find the catpkg than for you do it yourself. When specifying catpkg, you should have reasonable assurance that
        the category is valid and PMS-compliant as the caching mechanism we use does not perform validation checks for
        categories.

        This function is required for GLEP 42 compliance.

        @param catpkg: catpkg for which we want a list of repositories; we'll get a list of all repos containing this
          catpkg; if None, return a list of all Repositories that contain a particular catpkg.
        @return: a list of repositories.
        """

        if catpkg is not None and self._better_cache is not None:
            return [repo.name for repo in self._better_cache[catpkg]]
        return self._ordered_repo_name_list

    def getMissingRepoNames(self):
        """
        Returns a list of repository paths that lack profiles/repo_name.
        """
        return self.settings.repositories.missing_repo_names

    def getIgnoredRepos(self):
        """
        Returns a list of repository paths that have been ignored, because
        another repo with the same name exists.
        """
        return self.settings.repositories.ignored_repos

    def findname2(self, mycpv, mytree=None, myrepo=None):
        """
        Returns the location of the CPV, and what overlay it was in.
        Searches overlays first, then PORTDIR; this allows us to return the first
        matching file.  As opposed to starting in portdir and then doing overlays
        second, we would have to exhaustively search the overlays until we found
        the file we wanted.
        If myrepo is not None it will find packages from this repository(overlay)
        """
        if not mycpv:
            return (None, 0)

        if myrepo is not None:
            mytree = self.treemap.get(myrepo)
            if mytree is None:
                return (None, 0)
        elif mytree is not None:
            # myrepo enables cached results when available
            myrepo = self.repositories.location_map.get(mytree)

        mysplit = mycpv.split("/")
        psplit = pkgsplit(mysplit[1])
        if psplit is None or len(mysplit) != 2:
            raise InvalidPackageName(mycpv)

        try:
            cp = mycpv.cp
        except AttributeError:
            cp = mysplit[0] + "/" + psplit[0]

        if self._better_cache is None:
            if mytree:
                mytrees = [mytree]
            else:
                mytrees = reversed(self.porttrees)
        else:
            try:
                repos = self._better_cache[cp]
            except KeyError:
                return (None, 0)

            mytrees = []
            for repo in repos:
                if mytree is not None and mytree != repo.location:
                    continue
                mytrees.append(repo.location)

        # For optimal performace in this hot spot, we do manual unicode
        # handling here instead of using the wrapped os module.
        encoding = _encodings["fs"]
        errors = "strict"

        relative_path = (
            mysplit[0] + _os.sep + psplit[0] + _os.sep + mysplit[1] + ".ebuild"
        )

        # There is no need to access the filesystem when the package
        # comes from this db and the package repo attribute corresponds
        # to the desired repo, since the file was previously found by
        # the cp_list method.
        if (
            myrepo is not None
            and myrepo == getattr(mycpv, "repo", None)
            and self is getattr(mycpv, "_db", None)
        ):
            return (mytree + _os.sep + relative_path, mytree)

        for x in mytrees:
            filename = x + _os.sep + relative_path
            if _os.access(
                _unicode_encode(filename, encoding=encoding, errors=errors), _os.R_OK
            ):
                return (filename, x)
        return (None, 0)

    def _write_cache(self, cpv, repo_path, metadata, ebuild_hash):

        try:
            cache = self.auxdb[repo_path]
            chf = cache.validation_chf
            metadata["_%s_" % chf] = getattr(ebuild_hash, chf)
        except CacheError:
            # Normally this shouldn't happen, so we'll show
            # a traceback for debugging purposes.
            traceback.print_exc()
            cache = None

        if cache is not None:
            try:
                cache[cpv] = metadata
            except CacheError:
                # Normally this shouldn't happen, so we'll show
                # a traceback for debugging purposes.
                traceback.print_exc()

    def _pull_valid_cache(self, cpv, ebuild_path, repo_path):
        try:
            ebuild_hash = eclass_cache.hashed_path(ebuild_path)
            # snag mtime since we use it later, and to trigger stat failure
            # if it doesn't exist
            ebuild_hash.mtime
        except FileNotFound:
            writemsg(
                _("!!! aux_get(): ebuild for " "'%s' does not exist at:\n") % (cpv,),
                noiselevel=-1,
            )
            writemsg("!!!            %s\n" % ebuild_path, noiselevel=-1)
            raise PortageKeyError(cpv)

        # Pull pre-generated metadata from the metadata/cache/
        # directory if it exists and is valid, otherwise fall
        # back to the normal writable cache.
        auxdbs = []
        pregen_auxdb = self._pregen_auxdb.get(repo_path)
        if pregen_auxdb is not None:
            auxdbs.append(pregen_auxdb)
        ro_auxdb = self._ro_auxdb.get(repo_path)
        if ro_auxdb is not None:
            auxdbs.append(ro_auxdb)
        auxdbs.append(self.auxdb[repo_path])
        eclass_db = self.repositories.get_repo_for_location(repo_path).eclass_db

        for auxdb in auxdbs:
            try:
                metadata = auxdb[cpv]
            except KeyError:
                continue
            except CacheError:
                if not auxdb.readonly:
                    try:
                        del auxdb[cpv]
                    except (KeyError, CacheError):
                        pass
                continue
            eapi = metadata.get("EAPI", "").strip()
            if not eapi:
                eapi = "0"
                metadata["EAPI"] = eapi
            if not eapi_is_supported(eapi):
                # Since we're supposed to be able to efficiently obtain the
                # EAPI from _parse_eapi_ebuild_head, we disregard cache entries
                # for unsupported EAPIs.
                continue
            if auxdb.validate_entry(metadata, ebuild_hash, eclass_db):
                break
        else:
            metadata = None

        return (metadata, ebuild_hash)

    def aux_get(self, mycpv, mylist, mytree=None, myrepo=None):
        "stub code for returning auxilliary db information, such as SLOT, DEPEND, etc."
        'input: "sys-apps/foo-1.0",["SLOT","DEPEND","HOMEPAGE"]'
        'return: ["0",">=sys-libs/bar-1.0","http://www.foo.com"] or raise PortageKeyError if error'
        # For external API consumers, self._event_loop returns a new event
        # loop on each access, so a local reference is needed in order
        # to avoid instantiating more than one.
        loop = self._event_loop
        return loop.run_until_complete(
            self.async_aux_get(mycpv, mylist, mytree=mytree, myrepo=myrepo, loop=loop)
        )

    def async_aux_get(self, mycpv, mylist, mytree=None, myrepo=None, loop=None):
        """
        Asynchronous form form of aux_get.

        @param mycpv: cpv for an ebuild
        @type mycpv: str
        @param mylist: list of metadata keys
        @type mylist: list
        @param mytree: The canonical path of the tree in which the ebuild
                is located, or None for automatic lookup
        @type mytree: str
        @param myrepo: name of the repo in which the ebuild is located,
                or None for automatic lookup
        @type myrepo: str
        @param loop: event loop (defaults to global event loop)
        @type loop: EventLoop
        @return: list of metadata values
        @rtype: asyncio.Future (or compatible)
        """
        # Don't default to self._event_loop here, since that creates a
        # local event loop for thread safety, and that could easily lead
        # to simultaneous instantiation of multiple event loops here.
        # Callers of this method certainly want the same event loop to
        # be used for all calls.
        loop = asyncio._wrap_loop(loop)
        future = loop.create_future()
        cache_me = False
        if myrepo is not None:
            mytree = self.treemap.get(myrepo)
            if mytree is None:
                future.set_exception(PortageKeyError(myrepo))
                return future

        if (
            mytree is not None
            and len(self.porttrees) == 1
            and mytree == self.porttrees[0]
        ):
            # mytree matches our only tree, so it's safe to
            # ignore mytree and cache the result
            mytree = None
            myrepo = None

        if mytree is None:
            cache_me = True
        if mytree is None and not self._known_keys.intersection(mylist).difference(
            self._aux_cache_keys
        ):
            aux_cache = self._aux_cache.get(mycpv)
            if aux_cache is not None:
                future.set_result([aux_cache.get(x, "") for x in mylist])
                return future
            cache_me = True

        try:
            cat, pkg = mycpv.split("/", 1)
        except ValueError:
            # Missing slash. Can't find ebuild so raise PortageKeyError.
            future.set_exception(PortageKeyError(mycpv))
            return future

        myebuild, mylocation = self.findname2(mycpv, mytree)

        if not myebuild:
            writemsg(
                "!!! aux_get(): %s\n" % _("ebuild not found for '%s'") % mycpv,
                noiselevel=1,
            )
            future.set_exception(PortageKeyError(mycpv))
            return future

        mydata, ebuild_hash = self._pull_valid_cache(mycpv, myebuild, mylocation)

        if mydata is not None:
            self._aux_get_return(
                future,
                mycpv,
                mylist,
                myebuild,
                ebuild_hash,
                mydata,
                mylocation,
                cache_me,
                None,
            )
            return future

        if myebuild in self._broken_ebuilds:
            future.set_exception(PortageKeyError(mycpv))
            return future

        proc = EbuildMetadataPhase(
            cpv=mycpv,
            ebuild_hash=ebuild_hash,
            portdb=self,
            repo_path=mylocation,
            scheduler=loop,
            settings=self.doebuild_settings,
        )

        proc.addExitListener(
            functools.partial(
                self._aux_get_return,
                future,
                mycpv,
                mylist,
                myebuild,
                ebuild_hash,
                mydata,
                mylocation,
                cache_me,
            )
        )
        future.add_done_callback(functools.partial(self._aux_get_cancel, proc))
        proc.start()
        return future

    @staticmethod
    def _aux_get_cancel(proc, future):
        if future.cancelled() and proc.returncode is None:
            proc.cancel()

    def _aux_get_return(
        self,
        future,
        mycpv,
        mylist,
        myebuild,
        ebuild_hash,
        mydata,
        mylocation,
        cache_me,
        proc,
    ):
        if future.cancelled():
            return
        if proc is not None:
            if proc.returncode != os.EX_OK:
                self._broken_ebuilds.add(myebuild)
                future.set_exception(PortageKeyError(mycpv))
                return
            mydata = proc.metadata
        mydata["repository"] = self.repositories.get_name_for_location(mylocation)
        mydata["_mtime_"] = ebuild_hash.mtime
        eapi = mydata.get("EAPI")
        if not eapi:
            eapi = "0"
            mydata["EAPI"] = eapi
        if eapi_is_supported(eapi):
            mydata["INHERITED"] = " ".join(mydata.get("_eclasses_", []))

        # finally, we look at our internal cache entry and return the requested data.
        returnme = [mydata.get(x, "") for x in mylist]

        if cache_me and self.frozen:
            aux_cache = {}
            for x in self._aux_cache_keys:
                aux_cache[x] = mydata.get(x, "")
            self._aux_cache[mycpv] = aux_cache

        future.set_result(returnme)

    def getFetchMap(self, mypkg, useflags=None, mytree=None):
        """
        Get the SRC_URI metadata as a dict which maps each file name to a
        set of alternative URIs.

        @param mypkg: cpv for an ebuild
        @type mypkg: String
        @param useflags: a collection of enabled USE flags, for evaluation of
                conditionals
        @type useflags: set, or None to enable all conditionals
        @param mytree: The canonical path of the tree in which the ebuild
                is located, or None for automatic lookup
        @type mypkg: String
        @return: A dict which maps each file name to a set of alternative
                URIs.
        @rtype: dict
        """
        loop = self._event_loop
        return loop.run_until_complete(
            self.async_fetch_map(mypkg, useflags=useflags, mytree=mytree, loop=loop)
        )

    def async_fetch_map(self, mypkg, useflags=None, mytree=None, loop=None):
        """
        Asynchronous form of getFetchMap.

        @param mypkg: cpv for an ebuild
        @type mypkg: String
        @param useflags: a collection of enabled USE flags, for evaluation of
                conditionals
        @type useflags: set, or None to enable all conditionals
        @param mytree: The canonical path of the tree in which the ebuild
                is located, or None for automatic lookup
        @type mypkg: String
        @param loop: event loop (defaults to global event loop)
        @type loop: EventLoop
        @return: A future that results in a dict which maps each file name to
                a set of alternative URIs.
        @rtype: asyncio.Future (or compatible)
        """
        loop = asyncio._wrap_loop(loop)
        result = loop.create_future()

        def aux_get_done(aux_get_future):
            if result.cancelled():
                return
            if aux_get_future.exception() is not None:
                if isinstance(aux_get_future.exception(), PortageKeyError):
                    # Convert this to an InvalidDependString exception since
                    # callers already handle it.
                    result.set_exception(
                        portage.exception.InvalidDependString(
                            "getFetchMap(): aux_get() error reading "
                            + mypkg
                            + "; aborting."
                        )
                    )
                else:
                    result.set_exception(future.exception())
                return

            eapi, myuris = aux_get_future.result()

            if not eapi_is_supported(eapi):
                # Convert this to an InvalidDependString exception
                # since callers already handle it.
                result.set_exception(
                    portage.exception.InvalidDependString(
                        "getFetchMap(): '{}' has unsupported EAPI: '{}'".format(
                            mypkg, eapi
                        )
                    )
                )
                return

            try:
                result.set_result(
                    _parse_uri_map(
                        mypkg, {"EAPI": eapi, "SRC_URI": myuris}, use=useflags
                    )
                )
            except Exception as e:
                result.set_exception(e)

        aux_get_future = self.async_aux_get(
            mypkg, ["EAPI", "SRC_URI"], mytree=mytree, loop=loop
        )
        result.add_done_callback(
            lambda result: aux_get_future.cancel() if result.cancelled() else None
        )
        aux_get_future.add_done_callback(aux_get_done)
        return result

    def getfetchsizes(self, mypkg, useflags=None, debug=0, myrepo=None):
        # returns a filename:size dictionnary of remaining downloads
        myebuild, mytree = self.findname2(mypkg, myrepo=myrepo)
        if myebuild is None:
            raise AssertionError(_("ebuild not found for '%s'") % mypkg)
        pkgdir = os.path.dirname(myebuild)
        mf = self.repositories.get_repo_for_location(
            os.path.dirname(os.path.dirname(pkgdir))
        ).load_manifest(pkgdir, self.settings["DISTDIR"])
        checksums = mf.getDigests()
        if not checksums:
            if debug:
                writemsg(_("[empty/missing/bad digest]: %s\n") % (mypkg,))
            return {}
        filesdict = {}
        myfiles = self.getFetchMap(mypkg, useflags=useflags, mytree=mytree)
        # XXX: maybe this should be improved: take partial downloads
        # into account? check checksums?
        for myfile in myfiles:
            try:
                fetch_size = int(checksums[myfile]["size"])
            except (KeyError, ValueError):
                if debug:
                    writemsg(
                        _("[bad digest]: missing %(file)s for %(pkg)s\n")
                        % {"file": myfile, "pkg": mypkg}
                    )
                continue
            file_path = os.path.join(self.settings["DISTDIR"], myfile)
            mystat = None
            try:
                mystat = os.stat(file_path)
            except OSError:
                pass
            else:
                if mystat.st_size != fetch_size:
                    # Use file with _download_suffix instead.
                    mystat = None

            if mystat is None:
                try:
                    mystat = os.stat(file_path + _download_suffix)
                except OSError:
                    pass

            if mystat is None:
                existing_size = 0
                ro_distdirs = self.settings.get("PORTAGE_RO_DISTDIRS")
                if ro_distdirs is not None:
                    for x in shlex_split(ro_distdirs):
                        try:
                            mystat = os.stat(
                                portage.package.ebuild.fetch.get_mirror_url(
                                    x, myfile, self.settings
                                )
                            )
                        except OSError:
                            pass
                        else:
                            if mystat.st_size == fetch_size:
                                existing_size = fetch_size
                                break
            else:
                existing_size = mystat.st_size
            remaining_size = fetch_size - existing_size
            if remaining_size > 0:
                # Assume the download is resumable.
                filesdict[myfile] = remaining_size
            elif remaining_size < 0:
                # The existing file is too large and therefore corrupt.
                filesdict[myfile] = int(checksums[myfile]["size"])
        return filesdict

    def fetch_check(
        self, mypkg, useflags=None, mysettings=None, all=False, myrepo=None
    ):  # pylint: disable=redefined-builtin
        """
        TODO: account for PORTAGE_RO_DISTDIRS
        """
        if all:
            useflags = None
        elif useflags is None:
            if mysettings:
                useflags = mysettings["USE"].split()
        if myrepo is not None:
            mytree = self.treemap.get(myrepo)
            if mytree is None:
                return False
        else:
            mytree = None

        myfiles = self.getFetchMap(mypkg, useflags=useflags, mytree=mytree)
        myebuild = self.findname(mypkg, myrepo=myrepo)
        if myebuild is None:
            raise AssertionError(_("ebuild not found for '%s'") % mypkg)
        pkgdir = os.path.dirname(myebuild)
        mf = self.repositories.get_repo_for_location(
            os.path.dirname(os.path.dirname(pkgdir))
        )
        mf = mf.load_manifest(pkgdir, self.settings["DISTDIR"])
        mysums = mf.getDigests()

        failures = {}
        for x in myfiles:
            if not mysums or x not in mysums:
                ok = False
                reason = _("digest missing")
            else:
                try:
                    ok, reason = portage.checksum.verify_all(
                        os.path.join(self.settings["DISTDIR"], x), mysums[x]
                    )
                except FileNotFound as e:
                    ok = False
                    reason = _("File Not Found: '%s'") % (e,)
            if not ok:
                failures[x] = reason
        if failures:
            return False
        return True

    def cpv_exists(self, mykey, myrepo=None):
        "Tells us whether an actual ebuild exists on disk (no masking)"
        cps2 = mykey.split("/")
        cps = catpkgsplit(mykey, silent=0)
        if not cps:
            # invalid cat/pkg-v
            return 0
        if self.findname(cps[0] + "/" + cps2[1], myrepo=myrepo):
            return 1
        return 0

    def cp_all(self, categories=None, trees=None, reverse=False, sort=True):
        """
        This returns a list of all keys in our tree or trees
        @param categories: optional list of categories to search or
                defaults to self.settings.categories
        @param trees: optional list of trees to search the categories in or
                defaults to self.porttrees
        @param reverse: reverse sort order (default is False)
        @param sort: return sorted results (default is True)
        @rtype list of [cat/pkg,...]
        """
        d = {}
        if categories is None:
            categories = self.settings.categories
        if trees is None:
            trees = self.porttrees
        for x in categories:
            for oroot in trees:
                for y in listdir(
                    oroot + "/" + x, EmptyOnError=1, ignorecvs=1, dirsonly=1
                ):
                    try:
                        atom = Atom("{}/{}".format(x, y))
                    except InvalidAtom:
                        continue
                    if atom != atom.cp:
                        continue
                    d[atom.cp] = None
        l = list(d)
        if sort:
            l.sort(reverse=reverse)
        return l

    def cp_list(self, mycp, use_cache=1, mytree=None):
        # NOTE: Cache can be safely shared with the match cache, since the
        # match cache uses the result from dep_expand for the cache_key.
        if (
            self.frozen
            and mytree is not None
            and len(self.porttrees) == 1
            and mytree == self.porttrees[0]
        ):
            # mytree matches our only tree, so it's safe to
            # ignore mytree and cache the result
            mytree = None

        if self.frozen and mytree is None:
            cachelist = self.xcache["cp-list"].get(mycp)
            if cachelist is not None:
                # Try to propagate this to the match-all cache here for
                # repoman since he uses separate match-all caches for each
                # profile (due to differences in _get_implicit_iuse).
                self.xcache["match-all"][(mycp, mycp)] = cachelist
                return cachelist[:]
        mysplit = mycp.split("/")
        invalid_category = mysplit[0] not in self._categories
        # Process repos in ascending order by repo.priority, so that
        # stable sort by version produces results ordered by
        # (pkg.version, repo.priority).
        if mytree is not None:
            if isinstance(mytree, str):
                repos = [self.repositories.get_repo_for_location(mytree)]
            else:
                # assume it's iterable
                repos = [
                    self.repositories.get_repo_for_location(location)
                    for location in mytree
                ]
        elif self._better_cache is None:
            repos = self._porttrees_repos.values()
        else:
            repos = [
                repo
                for repo in reversed(self._better_cache[mycp])
                if repo.name in self._porttrees_repos
            ]
        mylist = []
        for repo in repos:
            oroot = repo.location
            try:
                file_list = os.listdir(os.path.join(oroot, mycp))
            except OSError:
                continue
            for x in file_list:
                pf = None
                if x[-7:] == ".ebuild":
                    pf = x[:-7]

                if pf is not None:
                    ps = pkgsplit(pf)
                    if not ps:
                        writemsg(
                            _("\nInvalid ebuild name: %s\n")
                            % os.path.join(oroot, mycp, x),
                            noiselevel=-1,
                        )
                        continue
                    if ps[0] != mysplit[1]:
                        writemsg(
                            _("\nInvalid ebuild name: %s\n")
                            % os.path.join(oroot, mycp, x),
                            noiselevel=-1,
                        )
                        continue
                    ver_match = ver_regexp.match("-".join(ps[1:]))
                    if ver_match is None or not ver_match.groups():
                        writemsg(
                            _("\nInvalid ebuild version: %s\n")
                            % os.path.join(oroot, mycp, x),
                            noiselevel=-1,
                        )
                        continue
                    mylist.append(
                        _pkg_str(mysplit[0] + "/" + pf, db=self, repo=repo.name)
                    )
        if invalid_category and mylist:
            writemsg(
                _(
                    "\n!!! '%s' has a category that is not listed in "
                    "%setc/portage/categories\n"
                )
                % (mycp, self.settings["PORTAGE_CONFIGROOT"]),
                noiselevel=-1,
            )
            mylist = []
        # Always sort in ascending order here since it's handy and
        # the result can be easily cached and reused. Since mylist
        # is initially in ascending order by repo.priority, stable
        # sort by version produces results in ascending order by
        # (pkg.version, repo.priority).
        self._cpv_sort_ascending(mylist)
        if self.frozen and mytree is None:
            cachelist = mylist[:]
            self.xcache["cp-list"][mycp] = cachelist
            self.xcache["match-all"][(mycp, mycp)] = cachelist
        return mylist

    def freeze(self):
        for x in (
            "bestmatch-visible",
            "cp-list",
            "match-all",
            "match-all-cpv-only",
            "match-visible",
            "minimum-all",
            "minimum-all-ignore-profile",
            "minimum-visible",
        ):
            self.xcache[x] = {}
        self.frozen = 1
        self._better_cache = _better_cache(self.repositories)

    def melt(self):
        self.xcache = {}
        self._aux_cache = {}
        self._better_cache = None
        self.frozen = 0

    def xmatch(
        self,
        level,
        origdep,
        mydep=DeprecationWarning,
        mykey=DeprecationWarning,
        mylist=DeprecationWarning,
    ):
        """
        Caching match function.

        @param level: xmatch level (bestmatch-visible, match-all-cpv-only
                match-allmatch-visible, minimum-all, minimum-all-ignore-profile,
                or minimum-visible)
        @type level: str
        @param origdep: dependency to match (may omit category)
        @type origdep: portage.dep.Atom or str
        @return: match result(s)
        @rtype: _pkg_str or list of _pkg_str (depends on level)
        """
        if level == "list-visible":
            level = "match-visible"
            warnings.warn(
                "The 'list-visible' mode of "
                "portage.dbapi.porttree.portdbapi.xmatch "
                "has been renamed to match-visible",
                DeprecationWarning,
                stacklevel=2,
            )

        if mydep is not DeprecationWarning:
            warnings.warn(
                "The 'mydep' parameter of "
                "portage.dbapi.porttree.portdbapi.xmatch"
                " is deprecated and ignored",
                DeprecationWarning,
                stacklevel=2,
            )

        loop = self._event_loop
        return loop.run_until_complete(self.async_xmatch(level, origdep, loop=loop))

    async def async_xmatch(self, level, origdep, loop=None):
        """
        Asynchronous form of xmatch.

        @param level: xmatch level (bestmatch-visible, match-all-cpv-only
                match-allmatch-visible, minimum-all, minimum-all-ignore-profile,
                or minimum-visible)
        @type level: str
        @param origdep: dependency to match (may omit category)
        @type origdep: portage.dep.Atom or str
        @param loop: event loop (defaults to global event loop)
        @type loop: EventLoop
        @return: match result(s)
        @rtype: asyncio.Future (or compatible), which results in a _pkg_str
                or list of _pkg_str (depends on level)
        """
        mydep = dep_expand(origdep, mydb=self, settings=self.settings)
        mykey = mydep.cp

        # if no updates are being made to the tree, we can consult our xcache...
        cache_key = None
        if self.frozen:
            cache_key = (mydep, mydep.unevaluated_atom)
            try:
                return self.xcache[level][cache_key][:]
            except KeyError:
                pass

        loop = asyncio._wrap_loop(loop)
        myval = None
        mytree = None
        if mydep.repo is not None:
            mytree = self.treemap.get(mydep.repo)
            if mytree is None:
                if level.startswith("match-"):
                    myval = []
                else:
                    myval = ""

        if myval is not None:
            # Unknown repo, empty result.
            pass
        elif level == "match-all-cpv-only":
            # match *all* packages, only against the cpv, in order
            # to bypass unnecessary cache access for things like IUSE
            # and SLOT.
            if mydep == mykey:
                # Share cache with match-all/cp_list when the result is the
                # same. Note that this requires that mydep.repo is None and
                # thus mytree is also None.
                level = "match-all"
                myval = self.cp_list(mykey, mytree=mytree)
            else:
                myval = match_from_list(mydep, self.cp_list(mykey, mytree=mytree))

        elif level in (
            "bestmatch-visible",
            "match-all",
            "match-visible",
            "minimum-all",
            "minimum-all-ignore-profile",
            "minimum-visible",
        ):
            # Find the minimum matching visible version. This is optimized to
            # minimize the number of metadata accesses (improves performance
            # especially in cases where metadata needs to be generated).
            if mydep == mykey:
                mylist = self.cp_list(mykey, mytree=mytree)
            else:
                mylist = match_from_list(mydep, self.cp_list(mykey, mytree=mytree))

            ignore_profile = level in ("minimum-all-ignore-profile",)
            visibility_filter = level not in (
                "match-all",
                "minimum-all",
                "minimum-all-ignore-profile",
            )
            single_match = level not in ("match-all", "match-visible")
            myval = []
            aux_keys = list(self._aux_cache_keys)
            if level == "bestmatch-visible":
                iterfunc = reversed
            else:
                iterfunc = iter

            for cpv in iterfunc(mylist):
                try:
                    metadata = dict(
                        zip(
                            aux_keys,
                            (
                                await self.async_aux_get(
                                    cpv, aux_keys, myrepo=cpv.repo, loop=loop
                                )
                            ),
                        )
                    )
                except KeyError:
                    # ebuild not in this repo, or masked by corruption
                    continue

                try:
                    pkg_str = _pkg_str(
                        cpv, metadata=metadata, settings=self.settings, db=self
                    )
                except InvalidData:
                    continue

                if visibility_filter and not self._visible(pkg_str, metadata):
                    continue

                if mydep.slot is not None and not _match_slot(mydep, pkg_str):
                    continue

                if mydep.unevaluated_atom.use is not None and not self._match_use(
                    mydep, pkg_str, metadata, ignore_profile=ignore_profile
                ):
                    continue

                myval.append(pkg_str)
                if single_match:
                    break

            if single_match:
                if myval:
                    myval = myval[0]
                else:
                    myval = ""

        else:
            raise AssertionError("Invalid level argument: '%s'" % level)

        if self.frozen:
            xcache_this_level = self.xcache.get(level)
            if xcache_this_level is not None:
                xcache_this_level[cache_key] = myval
                if not isinstance(myval, _pkg_str):
                    myval = myval[:]

        return myval

    def match(self, mydep, use_cache=1):
        return self.xmatch("match-visible", mydep)

    def gvisible(self, mylist):
        warnings.warn(
            "The 'gvisible' method of "
            "portage.dbapi.porttree.portdbapi "
            "is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self._iter_visible(iter(mylist)))

    def visible(self, cpv_iter):
        warnings.warn(
            "The 'visible' method of "
            "portage.dbapi.porttree.portdbapi "
            "is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        if cpv_iter is None:
            return []
        return list(self._iter_visible(iter(cpv_iter)))

    def _iter_visible(self, cpv_iter, myrepo=None):
        """
        Return a new list containing only visible packages.
        """
        aux_keys = list(self._aux_cache_keys)
        metadata = {}

        if myrepo is not None:
            repos = [myrepo]
        else:
            # We iterate over self.porttrees, since it's common to
            # tweak this attribute in order to adjust match behavior.
            repos = []
            for tree in reversed(self.porttrees):
                repos.append(self.repositories.get_name_for_location(tree))

        for mycpv in cpv_iter:
            for repo in repos:
                metadata.clear()
                try:
                    metadata.update(
                        zip(aux_keys, self.aux_get(mycpv, aux_keys, myrepo=repo))
                    )
                except KeyError:
                    continue
                except PortageException as e:
                    writemsg(
                        "!!! Error: aux_get('{}', {})\n".format(mycpv, aux_keys),
                        noiselevel=-1,
                    )
                    writemsg("!!! {}\n".format(e), noiselevel=-1)
                    del e
                    continue

                if not self._visible(mycpv, metadata):
                    continue

                yield mycpv
                # only yield a given cpv once
                break

    def _visible(self, cpv, metadata):
        eapi = metadata["EAPI"]
        if not eapi_is_supported(eapi):
            return False
        if _eapi_is_deprecated(eapi):
            return False
        if not metadata["SLOT"]:
            return False

        settings = self.settings
        if settings._getMaskAtom(cpv, metadata):
            return False
        if settings._getMissingKeywords(cpv, metadata):
            return False
        if settings.local_config:
            metadata["CHOST"] = settings.get("CHOST", "")
            if not settings._accept_chost(cpv, metadata):
                return False
            metadata["USE"] = ""
            if "?" in metadata["LICENSE"] or "?" in metadata["PROPERTIES"]:
                self.doebuild_settings.setcpv(cpv, mydb=metadata)
                metadata["USE"] = self.doebuild_settings["PORTAGE_USE"]
            try:
                if settings._getMissingLicenses(cpv, metadata):
                    return False
                if settings._getMissingProperties(cpv, metadata):
                    return False
                if settings._getMissingRestrict(cpv, metadata):
                    return False
            except InvalidDependString:
                return False

        return True


class portagetree:
    def __init__(
        self, root=DeprecationWarning, virtual=DeprecationWarning, settings=None
    ):
        """
        Constructor for a PortageTree

        @param root: deprecated, defaults to settings['ROOT']
        @type root: String/Path
        @param virtual: UNUSED
        @type virtual: No Idea
        @param settings: Portage Configuration object (portage.settings)
        @type settings: Instance of portage.config
        """

        if settings is None:
            settings = portage.settings
        self.settings = settings

        if root is not DeprecationWarning:
            warnings.warn(
                "The root parameter of the "
                + "portage.dbapi.porttree.portagetree"
                + " constructor is now unused. Use "
                + "settings['ROOT'] instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        if virtual is not DeprecationWarning:
            warnings.warn(
                "The 'virtual' parameter of the "
                "portage.dbapi.porttree.portagetree"
                " constructor is unused",
                DeprecationWarning,
                stacklevel=2,
            )

        self.__virtual = virtual
        self.dbapi = portdbapi(mysettings=settings)

    @property
    def portroot(self):
        """Deprecated. Use the portdbapi getRepositoryPath method instead."""
        warnings.warn(
            "The portroot attribute of "
            "portage.dbapi.porttree.portagetree is deprecated. Use the "
            "portdbapi getRepositoryPath method instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        return self.settings["PORTDIR"]

    @property
    def root(self):
        warnings.warn(
            "The root attribute of "
            + "portage.dbapi.porttree.portagetree"
            + " is deprecated. Use "
            + "settings['ROOT'] instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        return self.settings["ROOT"]

    @property
    def virtual(self):
        warnings.warn(
            "The 'virtual' attribute of "
            + "portage.dbapi.porttree.portagetree"
            + " is deprecated.",
            DeprecationWarning,
            stacklevel=3,
        )
        return self.__virtual

    def dep_bestmatch(self, mydep):
        "compatibility method"
        mymatch = self.dbapi.xmatch("bestmatch-visible", mydep)
        if mymatch is None:
            return ""
        return mymatch

    def dep_match(self, mydep):
        "compatibility method"
        mymatch = self.dbapi.xmatch("match-visible", mydep)
        if mymatch is None:
            return []
        return mymatch

    def exists_specific(self, cpv):
        return self.dbapi.cpv_exists(cpv)

    def getallnodes(self):
        """new behavior: these are all *unmasked* nodes.  There may or may not be available
        masked package for nodes in this nodes list."""
        return self.dbapi.cp_all()

    def getname(self, pkgname):
        """Deprecated. Use the portdbapi findname method instead."""
        warnings.warn(
            "The getname method of "
            "portage.dbapi.porttree.portagetree is deprecated. "
            "Use the portdbapi findname method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if not pkgname:
            return ""
        mysplit = pkgname.split("/")
        psplit = pkgsplit(mysplit[1])
        return "/".join([self.portroot, mysplit[0], psplit[0], mysplit[1]]) + ".ebuild"

    def getslot(self, mycatpkg):
        "Get a slot for a catpkg; assume it exists."
        myslot = ""
        try:
            myslot = self.dbapi._pkg_str(mycatpkg, None).slot
        except KeyError:
            pass
        return myslot


class FetchlistDict(Mapping):
    """
    This provide a mapping interface to retrieve fetch lists. It's used
    to allow portage.manifest.Manifest to access fetch lists via a standard
    mapping interface rather than use the dbapi directly.
    """

    def __init__(self, pkgdir, settings, mydbapi):
        """pkgdir is a directory containing ebuilds and settings is passed into
        portdbapi.getfetchlist for __getitem__ calls."""
        self.pkgdir = pkgdir
        self.cp = os.sep.join(pkgdir.split(os.sep)[-2:])
        self.settings = settings
        self.mytree = os.path.realpath(os.path.dirname(os.path.dirname(pkgdir)))
        self.portdb = mydbapi

    def __getitem__(self, pkg_key):
        """Returns the complete fetch list for a given package."""
        return list(self.portdb.getFetchMap(pkg_key, mytree=self.mytree))

    def __contains__(self, cpv):
        return cpv in self.__iter__()

    def has_key(self, pkg_key):
        """Returns true if the given package exists within pkgdir."""
        warnings.warn(
            "portage.dbapi.porttree.FetchlistDict.has_key() is "
            "deprecated, use the 'in' operator instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return pkg_key in self

    def __iter__(self):
        return iter(self.portdb.cp_list(self.cp, mytree=self.mytree))

    def __len__(self):
        """This needs to be implemented in order to avoid
        infinite recursion in some cases."""
        return len(self.portdb.cp_list(self.cp, mytree=self.mytree))

    keys = __iter__


def _async_manifest_fetchlist(
    portdb, repo_config, cp, cpv_list=None, max_jobs=None, max_load=None, loop=None
):
    """
    Asynchronous form of FetchlistDict, with max_jobs and max_load
    parameters in order to control async_aux_get concurrency.

    @param portdb: portdbapi instance
    @type portdb: portdbapi
    @param repo_config: repository configuration for a Manifest
    @type repo_config: RepoConfig
    @param cp: cp for a Manifest
    @type cp: str
    @param cpv_list: list of ebuild cpv values for a Manifest
    @type cpv_list: list
    @param max_jobs: max number of futures to process concurrently (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_jobs: int
    @param max_load: max load allowed when scheduling a new future,
            otherwise schedule no more than 1 future at a time (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_load: int or float
    @param loop: event loop
    @type loop: EventLoop
    @return: a Future resulting in a Mapping compatible with FetchlistDict
    @rtype: asyncio.Future (or compatible)
    """
    loop = asyncio._wrap_loop(loop)
    result = loop.create_future()
    cpv_list = (
        portdb.cp_list(cp, mytree=repo_config.location)
        if cpv_list is None
        else cpv_list
    )

    def gather_done(gather_result):
        # All exceptions must be consumed from gather_result before this
        # function returns, in order to avoid triggering the event loop's
        # exception handler.
        e = None
        if not gather_result.cancelled():
            for future in gather_result.result():
                if (
                    future.done()
                    and not future.cancelled()
                    and future.exception() is not None
                ):
                    e = future.exception()

        if result.cancelled():
            return
        if e is None:
            result.set_result(
                {k: list(v.result()) for k, v in zip(cpv_list, gather_result.result())}
            )
        else:
            result.set_exception(e)

    gather_result = iter_gather(
        # Use a generator expression for lazy evaluation, so that iter_gather
        # controls the number of concurrent async_fetch_map calls.
        (
            portdb.async_fetch_map(cpv, mytree=repo_config.location, loop=loop)
            for cpv in cpv_list
        ),
        max_jobs=max_jobs,
        max_load=max_load,
        loop=loop,
    )

    gather_result.add_done_callback(gather_done)
    result.add_done_callback(
        lambda result: gather_result.cancel() if result.cancelled() else None
    )

    return result


def _parse_uri_map(cpv, metadata, use=None):

    myuris = use_reduce(
        metadata.get("SRC_URI", ""),
        uselist=use,
        matchall=(use is None),
        is_src_uri=True,
        eapi=metadata["EAPI"],
    )

    uri_map = OrderedDict()

    myuris.reverse()
    while myuris:
        uri = myuris.pop()
        if myuris and myuris[-1] == "->":
            myuris.pop()
            distfile = myuris.pop()
        else:
            distfile = os.path.basename(uri)
            if not distfile:
                raise portage.exception.InvalidDependString(
                    ("getFetchMap(): '%s' SRC_URI has no file " + "name: '%s'")
                    % (cpv, uri)
                )

        uri_set = uri_map.get(distfile)
        if uri_set is None:
            # Use OrderedDict to preserve order from SRC_URI
            # while ensuring uniqueness.
            uri_set = OrderedDict()
            uri_map[distfile] = uri_set

        # SRC_URI may contain a file name with no scheme, and in
        # this case it does not belong in uri_set.
        if urlparse(uri).scheme:
            uri_set[uri] = True

    # Convert OrderedDicts to tuples.
    for k, v in uri_map.items():
        uri_map[k] = tuple(v)

    return uri_map



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["MtimeDB"]

import copy

try:
    import cPickle as pickle
except ImportError:
    import pickle

import errno
import io
import json

import portage
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.data import portage_gid, uid
from portage.localization import _
from portage.util import apply_secpass_permissions, atomic_ofstream, writemsg


_MTIMEDBKEYS = {
    "info",
    "ldpath",
    "resume",
    "resume_backup",
    "starttime",
    "updates",
    "version",
}


class MtimeDB(dict):
    """The MtimeDB class is used to interact with a file storing the
    current resume lists.
    It is a subclass of ``dict`` and it reads from/writes to JSON, by
    default, althouth it can be configured to use ``pickle``.
    """

    # JSON read support has been available since portage-2.1.10.49.
    _json_write = True

    _json_write_opts = {"ensure_ascii": False, "indent": "\t", "sort_keys": True}

    def __init__(self, filename):
        dict.__init__(self)
        self.filename = filename
        self._load(filename)

    @property
    def is_readonly(self):
        if self.filename is None:
            return True
        else:
            return False

    def make_readonly(self):
        self.filename = None

    def _load(self, filename):
        f = None
        content = None
        try:
            f = open(_unicode_encode(filename), "rb")
            content = f.read()
        except OSError as e:
            if getattr(e, "errno", None) in (errno.ENOENT, errno.EACCES):
                pass
            else:
                writemsg(_(f"!!! Error loading '{filename}': {e}\n"), noiselevel=-1)
        finally:
            if f is not None:
                f.close()

        d = {}
        if content:
            try:
                d = json.loads(
                    _unicode_decode(
                        content, encoding=_encodings["repo.content"], errors="strict"
                    )
                )
            except SystemExit:
                raise
            except Exception as e:
                try:
                    mypickle = pickle.Unpickler(io.BytesIO(content))
                    try:
                        mypickle.find_global = None
                    except AttributeError:
                        # Python >=3
                        pass
                    d = mypickle.load()
                except SystemExit:
                    raise
                except Exception:
                    writemsg(_(f"!!! Error loading '{filename}': {e}\n"), noiselevel=-1)

        if "old" in d:
            d["updates"] = d["old"]
            del d["old"]
        if "cur" in d:
            del d["cur"]

        d.setdefault("starttime", 0)
        d.setdefault("version", "")
        for k in ("info", "ldpath", "updates"):
            d.setdefault(k, {})

        for k in set(d.keys()) - _MTIMEDBKEYS:
            writemsg(_(f"Deleting invalid mtimedb key: {k}\n"))
            del d[k]
        self.update(d)
        self._clean_data = copy.deepcopy(d)

    def commit(self):
        if self.is_readonly:
            return
        d = {}
        d.update(self)
        # Only commit if the internal state has changed.
        if d != self._clean_data:
            self.__write_to_disk(d)

    def __write_to_disk(self, d):
        """Private method used by the ``commit`` method."""
        d["version"] = str(portage.VERSION)
        try:
            f = atomic_ofstream(self.filename, mode="wb")
        except OSError:
            pass
        else:
            if self._json_write:
                f.write(
                    _unicode_encode(
                        json.dumps(d, **self._json_write_opts),
                        encoding=_encodings["repo.content"],
                        errors="strict",
                    )
                )
            else:
                pickle.dump(d, f, protocol=2)
            f.close()
            apply_secpass_permissions(
                self.filename, uid=uid, gid=portage_gid, mode=0o644
            )
            self._clean_data = copy.deepcopy(d)



# Copyright 2007-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging

from portage import os
from portage.repository.config import allow_profile_repo_deps
from portage.util import grabfile_package, stack_lists
from portage._sets.base import PackageSet
from portage._sets import get_boolean
from portage.util import writemsg_level

__all__ = ["PackagesSystemSet"]


class PackagesSystemSet(PackageSet):
    _operations = ["merge"]

    def __init__(self, profiles, debug=False):
        super().__init__(allow_repo=any(allow_profile_repo_deps(x) for x in profiles))
        self._profiles = profiles
        self._debug = debug
        if profiles:
            desc_profile = profiles[-1]
            if desc_profile.user_config and len(profiles) > 1:
                desc_profile = profiles[-2]
            description = desc_profile.location
        else:
            description = None
        self.description = "System packages for profile %s" % description

    def load(self):
        debug = self._debug
        if debug:
            writemsg_level(
                "\nPackagesSystemSet: profiles: {}\n".format(self._profiles),
                level=logging.DEBUG,
                noiselevel=-1,
            )

        mylist = [
            grabfile_package(
                os.path.join(x.location, "packages"),
                verify_eapi=True,
                eapi=x.eapi,
                eapi_default=None,
                allow_build_id=x.allow_build_id,
                allow_repo=allow_profile_repo_deps(x),
            )
            for x in self._profiles
        ]

        if debug:
            writemsg_level(
                "\nPackagesSystemSet: raw packages: {}\n".format(mylist),
                level=logging.DEBUG,
                noiselevel=-1,
            )

        mylist = stack_lists(mylist, incremental=1)

        if debug:
            writemsg_level(
                "\nPackagesSystemSet: stacked packages: {}\n".format(mylist),
                level=logging.DEBUG,
                noiselevel=-1,
            )

        self._setAtoms([x[1:] for x in mylist if x[0] == "*"])

    def singleBuilder(self, options, settings, trees):
        debug = get_boolean(options, "debug", False)
        return PackagesSystemSet(
            settings._locations_manager.profiles_complex, debug=debug
        )

    singleBuilder = classmethod(singleBuilder)



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.util._ctypes import find_library, LoadLibrary
from portage.util._async.ForkProcess import ForkProcess


class SyncfsProcess(ForkProcess):
    """
    Isolate ctypes usage in a subprocess, in order to avoid
    potential problems with stale cached libraries as
    described in bug #448858, comment #14 (also see
    https://bugs.python.org/issue14597).
    """

    __slots__ = ("paths",)

    @staticmethod
    def _get_syncfs():

        filename = find_library("c")
        if filename is not None:
            library = LoadLibrary(filename)
            if library is not None:
                try:
                    return library.syncfs
                except AttributeError:
                    pass

        return None

    def _run(self):

        syncfs_failed = False
        syncfs = self._get_syncfs()

        if syncfs is not None:
            for path in self.paths:
                try:
                    fd = os.open(path, os.O_RDONLY)
                except OSError:
                    pass
                else:
                    try:
                        if syncfs(fd) != 0:
                            # Happens with PyPy (bug #446610)
                            syncfs_failed = True
                    finally:
                        os.close(fd)

        if syncfs is None or syncfs_failed:
            return 1
        return os.EX_OK



# Copyright 2012-2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import subprocess
import tempfile

from portage import os
from portage.tests import TestCase
from portage.util._async.PipeLogger import PipeLogger
from portage.util._async.PopenProcess import PopenProcess
from portage.util._eventloop.global_event_loop import global_event_loop
from _emerge.PipeReader import PipeReader


class PopenPipeTestCase(TestCase):
    """
    Test PopenProcess, which is historically useful for Jython support,
    since it uses the subprocess.Popen instead of os.fork().

    Portage does not currently support Jython, but re-introducing support
    in The Future (TM) may be possible.
    """

    _echo_cmd = "echo -n '%s'"

    def _testPipeReader(self, test_string):
        """
        Use a poll loop to read data from a pipe and assert that
        the data written to the pipe is identical to the data
        read from the pipe.
        """

        producer = PopenProcess(
            proc=subprocess.Popen(
                ["bash", "-c", self._echo_cmd % test_string],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            ),
            pipe_reader=PipeReader(),
            scheduler=global_event_loop(),
        )

        consumer = producer.pipe_reader
        consumer.input_files = {"producer": producer.proc.stdout}

        producer.start()
        producer.wait()

        self.assertEqual(producer.returncode, os.EX_OK)
        self.assertEqual(consumer.returncode, os.EX_OK)

        return consumer.getvalue().decode("ascii", "replace")

    def _testPipeLogger(self, test_string):

        producer = PopenProcess(
            proc=subprocess.Popen(
                ["bash", "-c", self._echo_cmd % test_string],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            ),
            scheduler=global_event_loop(),
        )

        fd, log_file_path = tempfile.mkstemp()
        try:

            consumer = PipeLogger(
                background=True,
                input_fd=producer.proc.stdout,
                log_file_path=log_file_path,
            )

            producer.pipe_reader = consumer

            producer.start()
            producer.wait()

            self.assertEqual(producer.returncode, os.EX_OK)
            self.assertEqual(consumer.returncode, os.EX_OK)

            with open(log_file_path, "rb") as f:
                content = f.read()

        finally:
            os.close(fd)
            os.unlink(log_file_path)

        return content.decode("ascii", "replace")

    def testPopenPipe(self):
        for x in (1, 2, 5, 6, 7, 8, 2**5, 2**10, 2**12, 2**13, 2**14):
            test_string = x * "a"
            output = self._testPipeReader(test_string)
            self.assertEqual(
                test_string, output, "x = {}, len(output) = {}".format(x, len(output))
            )

            output = self._testPipeLogger(test_string)
            self.assertEqual(
                test_string, output, "x = {}, len(output) = {}".format(x, len(output))
            )



# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

# Since python ebuilds remove the 'email' module when USE=build
# is enabled, use a local import so that
# portage.proxy.lazyimport._preload_portage_submodules()
# can load this module even though the 'email' module is missing.
# The elog mail modules won't work, but at least an ImportError
# won't cause portage to crash during stage builds. Since the
# 'smtlib' module imports the 'email' module, that's imported
# locally as well.

import sys

from portage import os
from portage import _unicode_decode, _unicode_encode
from portage.localization import _
import portage


def _force_ascii_if_necessary(s):
    # Force ascii encoding in order to avoid UnicodeEncodeError
    # from smtplib.sendmail with python3 (bug #291331).
    s = _unicode_encode(s, encoding="ascii", errors="backslashreplace")
    s = _unicode_decode(s, encoding="ascii", errors="replace")
    return s


def TextMessage(_text):
    from email.mime.text import MIMEText

    mimetext = MIMEText(_text)
    mimetext.set_charset("UTF-8")
    return mimetext


def create_message(sender, recipient, subject, body, attachments=None):

    from email.header import Header
    from email.mime.base import MIMEBase as BaseMessage
    from email.mime.multipart import MIMEMultipart as MultipartMessage
    from email.utils import formatdate

    if attachments is None:
        mymessage = TextMessage(body)
    else:
        mymessage = MultipartMessage()
        mymessage.attach(TextMessage(body))
        for x in attachments:
            if isinstance(x, BaseMessage):
                mymessage.attach(x)
            elif isinstance(x, str):
                mymessage.attach(TextMessage(x))
            else:
                raise portage.exception.PortageException(
                    _(f"Can't handle type of attachment: {type(x)}")
                )

    mymessage.set_unixfrom(sender)
    mymessage["To"] = recipient
    mymessage["From"] = sender

    # Use Header as a workaround so that long subject lines are wrapped
    # correctly by <=python-2.6 (gentoo bug #263370, python issue #1974).
    # Also, need to force ascii for python3, in order to avoid
    # UnicodeEncodeError with non-ascii characters:
    #  File "/usr/lib/python3.1/email/header.py", line 189, in __init__
    #    self.append(s, charset, errors)
    #  File "/usr/lib/python3.1/email/header.py", line 262, in append
    #    input_bytes = s.encode(input_charset, errors)
    # UnicodeEncodeError: 'ascii' codec can't encode characters in position 0-9: ordinal not in range(128)
    mymessage["Subject"] = Header(_force_ascii_if_necessary(subject))
    mymessage["Date"] = formatdate(localtime=True)

    return mymessage


def send_mail(mysettings, message):

    import smtplib

    mymailhost = "localhost"
    mymailport = 25
    mymailuser = ""
    mymailpasswd = ""
    myrecipient = "root@localhost"

    # Syntax for PORTAGE_ELOG_MAILURI (if defined):
    # address [[user:passwd@]mailserver[:port]]
    # where address:    recipient address
    #       user:       username for smtp auth (defaults to none)
    #       passwd:     password for smtp auth (defaults to none)
    #       mailserver: smtp server that should be used to deliver the mail (defaults to localhost)
    # 					alternatively this can also be the absolute path to a sendmail binary if you don't want to use smtp
    #       port:       port to use on the given smtp server (defaults to 25, values > 100000 indicate that starttls should be used on (port-100000))
    if " " in mysettings.get("PORTAGE_ELOG_MAILURI", ""):
        myrecipient, mymailuri = mysettings["PORTAGE_ELOG_MAILURI"].split()
        if "@" in mymailuri:
            myauthdata, myconndata = mymailuri.rsplit("@", 1)
            try:
                mymailuser, mymailpasswd = myauthdata.split(":")
            except ValueError:
                print(
                    _("!!! invalid SMTP AUTH configuration, trying unauthenticated ...")
                )
        else:
            myconndata = mymailuri
        if ":" in myconndata:
            mymailhost, mymailport = myconndata.split(":")
        else:
            mymailhost = myconndata
    else:
        myrecipient = mysettings.get("PORTAGE_ELOG_MAILURI", "")

    myfrom = message.get("From")

    # user wants to use a sendmail binary instead of smtp
    if mymailhost[0] == os.sep and os.path.exists(mymailhost):
        fd = os.popen(f"{mymailhost } -f {myfrom} {myrecipient}", "w")
        fd.write(_force_ascii_if_necessary(message.as_string()))
        if fd.close() is not None:
            sys.stderr.write(
                _(
                    f"!!! {mymailhost} returned with a non-zero exit code. This generally indicates an error.\n"
                )
            )
    else:
        try:
            if int(mymailport) > 100000:
                myconn = smtplib.SMTP(mymailhost, int(mymailport) - 100000)
                myconn.ehlo()
                if not myconn.has_extn("STARTTLS"):
                    raise portage.exception.PortageException(
                        _(
                            "!!! TLS support requested for logmail but not supported by server"
                        )
                    )
                myconn.starttls()
                myconn.ehlo()
            else:
                myconn = smtplib.SMTP(mymailhost, mymailport)
            if mymailuser != "" and mymailpasswd != "":
                myconn.login(mymailuser, mymailpasswd)

            message_str = _force_ascii_if_necessary(message.as_string())
            myconn.sendmail(myfrom, myrecipient, message_str)
            myconn.quit()
        except smtplib.SMTPException as e:
            raise portage.exception.PortageException(
                _(f"!!! An error occurred while trying to send logmail:\n{e}")
            )
        except OSError as e:
            raise portage.exception.PortageException(
                _(
                    f"!!! A network error occurred while trying to send logmail:\n{e}\nSure you configured PORTAGE_ELOG_MAILURI correctly?"
                )
            )



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import os

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "_emerge.PipeReader:PipeReader",
    "portage.util.futures:asyncio",
)


def _reader(input_file, loop=None):
    """
    Asynchronously read a binary input file, and close it when
    it reaches EOF.

    @param input_file: binary input file descriptor
    @type input_file: file or int
    @param loop: asyncio.AbstractEventLoop (or compatible)
    @type loop: event loop
    @return: bytes
    @rtype: asyncio.Future (or compatible)
    """
    loop = asyncio._wrap_loop(loop)
    future = loop.create_future()
    _Reader(future, input_file, loop)
    return future


class _Reader:
    def __init__(self, future, input_file, loop):
        self._future = future
        self._pipe_reader = PipeReader(
            input_files={"input_file": input_file}, scheduler=loop
        )

        self._future.add_done_callback(self._cancel_callback)
        self._pipe_reader.addExitListener(self._eof)
        self._pipe_reader.start()

    def _cancel_callback(self, future):
        if future.cancelled():
            self._cancel()

    def _eof(self, pipe_reader):
        self._pipe_reader = None
        self._future.set_result(pipe_reader.getvalue())

    def _cancel(self):
        if self._pipe_reader is not None and self._pipe_reader.poll() is None:
            self._pipe_reader.removeExitListener(self._eof)
            self._pipe_reader.cancel()
            self._pipe_reader = None


async def _writer(output_file, content, loop=DeprecationWarning):
    """
    Asynchronously write bytes to output file. The output file is
    assumed to be in non-blocking mode. If an EnvironmentError
    other than EAGAIN is encountered, which typically indicates that
    the other end of the pipe has closed, the error is raised.
    This function is a coroutine.

    @param output_file: output file
    @type output_file: file object
    @param content: content to write
    @type content: bytes
    @param loop: deprecated
    """
    loop = asyncio.get_event_loop()
    fd = output_file.fileno()
    while content:
        try:
            content = content[os.write(fd, content) :]
        except OSError as e:
            if e.errno != errno.EAGAIN:
                raise
            waiter = loop.create_future()
            loop.add_writer(fd, lambda: waiter.done() or waiter.set_result(None))
            try:
                await waiter
            finally:
                # The loop and output file may have been closed.
                if not loop.is_closed():
                    waiter.done() or waiter.cancel()
                    # Do not call remove_writer in cases where fd has
                    # been closed and then re-allocated to a concurrent
                    # coroutine as in bug 716636.
                    if not output_file.closed:
                        loop.remove_writer(fd)



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class RegularSlotChangeWithoutRevBumpTestCase(TestCase):
    def testRegularSlotChangeWithoutRevBumpTestCase(self):

        ebuilds = {
            "dev-libs/boost-1.52.0": {"SLOT": "0"},
            "app-office/libreoffice-4.0.0.2": {
                "EAPI": "5",
                "DEPEND": ">=dev-libs/boost-1.46:=",
                "RDEPEND": ">=dev-libs/boost-1.46:=",
            },
        }

        binpkgs = {
            "dev-libs/boost-1.52.0": {"SLOT": "1.52"},
        }

        installed = {
            "dev-libs/boost-1.52.0": {"SLOT": "1.52"},
        }

        world = []

        test_cases = (
            # Test that @__auto_slot_operator_replace_installed__
            # pulls in the available slot, even though it's
            # different from the installed slot (0 instead of 1.52).
            ResolverPlaygroundTestCase(
                ["app-office/libreoffice"],
                options={"--oneshot": True, "--usepkg": True},
                success=True,
                mergelist=["dev-libs/boost-1.52.0", "app-office/libreoffice-4.0.0.2"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# Copyright 2005-2021 Gentoo Authors
# Author(s): Brian Harring (ferringb@gentoo.org)
# License: GPL2

import errno
import re
import stat
from operator import attrgetter

import portage
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage.cache import cache_errors, flat_hash
import portage.eclass_cache
from portage.cache.template import reconstruct_eclasses
from portage.cache.mappings import ProtectedDict


# this is the old cache format, flat_list.  count maintained here.
magic_line_count = 22

# store the current key order *here*.


class database(flat_hash.database):
    complete_eclass_entries = False
    auxdbkey_order = (
        "DEPEND",
        "RDEPEND",
        "SLOT",
        "SRC_URI",
        "RESTRICT",
        "HOMEPAGE",
        "LICENSE",
        "DESCRIPTION",
        "KEYWORDS",
        "IDEPEND",
        "INHERITED",
        "IUSE",
        "REQUIRED_USE",
        "PDEPEND",
        "BDEPEND",
        "EAPI",
        "PROPERTIES",
        "DEFINED_PHASES",
    )

    autocommits = True
    serialize_eclasses = False

    _hashed_re = re.compile("^(\\w+)=([^\n]*)")

    def __init__(self, location, *args, **config):
        loc = location
        super().__init__(location, *args, **config)
        self.location = os.path.join(loc, "metadata", "cache")
        self.ec = None
        self.raise_stat_collision = False

    def _parse_data(self, data, cpv):
        _hashed_re_match = self._hashed_re.match
        d = {}

        for line in data:
            hashed = False
            hashed_match = _hashed_re_match(line)
            if hashed_match is None:
                d.clear()
                try:
                    for i, key in enumerate(self.auxdbkey_order):
                        d[key] = data[i]
                except IndexError:
                    pass
                break
            else:
                d[hashed_match.group(1)] = hashed_match.group(2)

        if "_eclasses_" not in d:
            if "INHERITED" in d:
                if self.ec is None:
                    self.ec = portage.eclass_cache.cache(self.location[:-15])
                getter = attrgetter(self.validation_chf)
                try:
                    ec_data = self.ec.get_eclass_data(d["INHERITED"].split())
                    d["_eclasses_"] = {
                        k: (v.eclass_dir, getter(v)) for k, v in ec_data.items()
                    }
                except KeyError as e:
                    # INHERITED contains a non-existent eclass.
                    raise cache_errors.CacheCorruption(cpv, e)
            else:
                d["_eclasses_"] = {}
        elif isinstance(d["_eclasses_"], str):
            # We skip this if flat_hash.database._parse_data() was called above
            # because it calls reconstruct_eclasses() internally.
            d["_eclasses_"] = reconstruct_eclasses(None, d["_eclasses_"])

        return d

    def _setitem(self, cpv, values):
        if "_eclasses_" in values:
            values = ProtectedDict(values)
            values["INHERITED"] = " ".join(sorted(values["_eclasses_"]))

        new_content = []
        for k in self.auxdbkey_order:
            new_content.append(values.get(k, ""))
            new_content.append("\n")
        for i in range(magic_line_count - len(self.auxdbkey_order)):
            new_content.append("\n")
        new_content = "".join(new_content)
        new_content = _unicode_encode(
            new_content, _encodings["repo.content"], errors="backslashreplace"
        )

        new_fp = os.path.join(self.location, cpv)
        try:
            f = open(
                _unicode_encode(new_fp, encoding=_encodings["fs"], errors="strict"),
                "rb",
            )
        except OSError:
            pass
        else:
            try:
                try:
                    existing_st = os.fstat(f.fileno())
                    existing_content = f.read()
                finally:
                    f.close()
            except OSError:
                pass
            else:
                existing_mtime = existing_st[stat.ST_MTIME]
                if (
                    values["_mtime_"] == existing_mtime
                    and existing_content == new_content
                ):
                    return

                if (
                    self.raise_stat_collision
                    and values["_mtime_"] == existing_mtime
                    and len(new_content) == existing_st.st_size
                ):
                    raise cache_errors.StatCollision(
                        cpv, new_fp, existing_mtime, existing_st.st_size
                    )

        s = cpv.rfind("/")
        fp = os.path.join(
            self.location, cpv[:s], ".update.%i.%s" % (portage.getpid(), cpv[s + 1 :])
        )
        try:
            myf = open(
                _unicode_encode(fp, encoding=_encodings["fs"], errors="strict"), "wb"
            )
        except OSError as e:
            if errno.ENOENT == e.errno:
                try:
                    self._ensure_dirs(cpv)
                    myf = open(
                        _unicode_encode(fp, encoding=_encodings["fs"], errors="strict"),
                        "wb",
                    )
                except OSError as e:
                    raise cache_errors.CacheCorruption(cpv, e)
            else:
                raise cache_errors.CacheCorruption(cpv, e)

        try:
            myf.write(new_content)
        finally:
            myf.close()
        self._ensure_access(fp, mtime=values["_mtime_"])

        try:
            os.rename(fp, new_fp)
        except OSError as e:
            try:
                os.unlink(fp)
            except OSError:
                pass
            raise cache_errors.CacheCorruption(cpv, e)
